/*----------------------------------------*/
/* Base Benchmark Test                    */
/*            using BNCpack, MPFR and GMP */
/* Copyright (c) 2004- Tomonori Kouya     */
/*                                        */
/* Version 0.0 : Nov.13, 2004             */
/* Version 0.1 : Jan.18, 2006             */
/*----------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "bnc.h"

void usage(const char *commandname)
{
	printf("USAGE-> %s MinimamDigits(Binary) MaximumDigits [Maximum seconds]\n", commandname);
}


/* kind of arithmetic */
/* 0 ... multiplication */
/* 1 ... division */
/* 2 ... sin */
/* 3 ... cos */
/* 4 ... exp */
/* 5 ... log */
/* 6 ... acos */
/* 7 ... atan */
int basebenchloop_by_clock(int kind_of_arithmetic, unsigned long int start_prec, unsigned long int end_prec, unsigned long int prec_step, long int maxsec)
{
	long int prec;
	long int i, j, end_flag = 0;
	double iteration, etime, stime;
	mpf_t a, b, c;
	double da, db, dc;

	printf("(%6ld - %6ld bits) :  clock/op ,milli-clock/op ,micro-clock/op \n", start_prec, end_prec);
	for(prec = start_prec; prec <= end_prec; prec *= prec_step)
	{
		if(kind_of_arithmetic >= 0)
		{
			mpf_init2(a, prec);
			mpf_init2(b, prec);
			mpf_init2(c, prec);

			/* the same args as timings-mpfr.c */
			mpf_set_ui(a, 5UL); mpf_sqrt(a, a); // a = sqrt(5);
			mpf_set_ui(b, 3UL); mpf_sqrt(b, b); mpf_sub_ui(b, b, 1UL); // b = sqrt(3) - 1;
		}
		else
		{
			da = sqrt(5.0);
			db = sqrt(3.0) - 1.0;
			end_flag = 1;
		}

		// Original Version is timings-mpfr.c at www.mpfr.org
		switch(kind_of_arithmetic)
		{
			case -2: // Multiplication
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) dc = da * db;
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 200000.0);
				break;
			case 0: // Multiplication
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) mpf_mul(c, a, b);
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 200000.0);
				break;
			case 1: // Division
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) mpf_div(c, a, b);
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 200000.0);
				break;
#ifdef USE_MPFR
			case 2: // sin
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) mpf_sin(c, a);
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 200000.0);
				break;
			case 3: // cos
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) mpf_cos(c, a);
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 200000.0);
				break;
			case 4: // exp
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) mpf_exp(c, a);
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 200000.0);
				break;
			case 5: // log
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) mpf_ln(c, a);
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 20000.0);
				break;
			case 6: // acos
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) mpf_acos(c, a);
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 20000.0);
				break;
			case 7: // atan
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) mpf_atan(c, a);
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 20000.0);
				break;
#endif
			default:
				iteration = 1.0;
				do{
					stime = (double)clock();
					for(i = 0; i < iteration; i++) mpf_mul(c, a, b);
					etime = (double)clock() - stime;
					iteration *= 2.0;
				} while(etime < 20000.0);
				break;
		}

//		if(etime > maxsec)
//			end_flag = 1;

		etime /= iteration / 2;

		switch(kind_of_arithmetic)
		{
			case -2: // Multiplication(IEEE754)
				printf("x * y (    53)        : %10.3g(clock/op ), %10.3g(seconds/mul), %10.3g(milli-seconds)\n", prec, etime, etime / CLOCKS_PER_SEC, etime / CLOCKS_PER_SEC * 1000 * 1000);
				break;
			case 0: // Multiplication
				printf("x * y (%6ld)        : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 1: // Division
				printf("x / y (%6ld)        : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
#ifdef USE_MPFR
			case 2: // sin
				printf("sin(x)(%6ld)        : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 3: // cos
				printf("cos(x)(%6ld)        : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 4: // exp
				printf("exp(x)(%6ld)        : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 5: // log
				printf("log(x)(%6ld)        : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 6: // acos
				printf("acos(x)(%6ld)       : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 7: // atan
				printf("atan(x)(%6ld)       : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
#endif
			default: // Multiplication
				printf("x * y (%6ld)        : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
		}

		if(kind_of_arithmetic >= 0)
		{
			mpf_clear(a);
			mpf_clear(b);
			mpf_clear(c);
		}

		if(end_flag == 1)
			break;
	}

	return end_flag;
}

/* kind of arithmetic */
/* 0 ... multiplication */
/* 1 ... division */
/* 2 ... sin */
/* 3 ... cos */
/* 4 ... exp */
/* 5 ... log */
/* 6 ... acos */
/* 7 ... atan */
int basebenchloop(int kind_of_arithmetic, unsigned long int start_prec, unsigned long int end_prec, unsigned long int prec_step, long int maxsec)
{
	long int prec;
	long int i, j, end_flag = 0;
	double iteration, etime, stime;
	mpf_t a, b, c;

	printf("(%6ld - %6ld bits) :    sec/op , milli-sec/op , micro-sec/op \n", start_prec, end_prec);
	for(prec = start_prec; prec <= end_prec; prec *= prec_step)
	{
		mpf_init2(a, prec);
		mpf_init2(b, prec);
		mpf_init2(c, prec);

		/* the same args as timings-mpfr.c */
		mpf_set_ui(a, 5UL); mpf_sqrt(a, a); // a = sqrt(5);
		mpf_set_ui(b, 3UL); mpf_sqrt(b, b); mpf_sub_ui(b, b, 1UL); // b = sqrt(3) - 1;

		// Original Version is timings-mpfr.c at www.mpfr.org
		switch(kind_of_arithmetic)
		{
			case 0: // Multiplication
				iteration = 1.0;
				do{
					stime = get_secv();
					for(i = 0; i < iteration; i++) mpf_mul(c, a, b);
					etime = get_secv() - stime;
					iteration *= 2.0;
				} while(etime < 2);
				break;
			case 1: // Division
				iteration = 1.0;
				do{
					stime = get_secv();
					for(i = 0; i < iteration; i++) mpf_div(c, a, b);
					etime = get_secv() - stime;
					iteration *= 2.0;
				} while(etime < 2);
				break;
#ifdef USE_MPFR
			case 2: // sin
				iteration = 1.0;
				do{
					stime = get_secv();
					for(i = 0; i < iteration; i++) mpf_sin(c, a);
					etime = get_secv() - stime;
					iteration *= 2.0;
				} while(etime < 2);
				break;
			case 3: // cos
				iteration = 1.0;
				do{
					stime = get_secv();
					for(i = 0; i < iteration; i++) mpf_cos(c, a);
					etime = get_secv() - stime;
					iteration *= 2.0;
				} while(etime < 2);
				break;
			case 4: // exp
				iteration = 1.0;
				do{
					stime = get_secv();
					for(i = 0; i < iteration; i++) mpf_exp(c, a);
					etime = get_secv() - stime;
					iteration *= 2.0;
				} while(etime < 2);
				break;
			case 5: // log
				iteration = 1.0;
				do{
					stime = get_secv();
					for(i = 0; i < iteration; i++) mpf_ln(c, a);
					etime = get_secv() - stime;
					iteration *= 2.0;
				} while(etime < 2);
				break;
			case 6: // acos
				iteration = 1.0;
				do{
					stime = get_secv();
					for(i = 0; i < iteration; i++) mpf_acos(c, b);
					etime = get_secv() - stime;
					iteration *= 2.0;
				} while(etime < 2);
				break;
			case 7: // atan
				iteration = 1.0;
				do{
					stime = get_secv();
					for(i = 0; i < iteration; i++) mpf_atan(c, a);
					etime = get_secv() - stime;
					iteration *= 2.0;
				} while(etime < 2);
				break;
#endif
			default:
				iteration = 1.0;
				do{
					stime = get_secv();
					for(i = 0; i < iteration; i++) mpf_mul(c, a, b);
					etime = get_secv() - stime;
					iteration *= 2.0;
				} while(etime < 2);
				break;
		}

		if(etime > maxsec)
			end_flag = 1;

		etime /= iteration / 2;

		switch(kind_of_arithmetic)
		{
			case 0: // Multiplication
				printf("x * y (%6ld)         : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 1: // Division
				printf("x / y (%6ld)         : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
#ifdef USE_MPFR
			case 2: // sin
				printf("sin(x)(%6ld)         : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 3: // cos
				printf("cos(x)(%6ld)         : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 4: // exp
				printf("exp(x)(%6ld)         : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 5: // log
				printf("log(x)(%6ld)         : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 6: // acos
				printf("acos(x)(%6ld)        : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
			case 7: // atan
				printf("atan(x)(%6ld)        : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
#endif
			default: // Multiplication
				printf("x * y (%6ld)         : %10.3g,    %10.3g,    %10.3g\n", prec, etime, etime * 1000, (etime) * 1000 * 1000);
				break;
		}

		mpf_clear(a);
		mpf_clear(b);
		mpf_clear(c);

		if(end_flag == 1)
			break;
	}

	return end_flag;
}

int main(int argc, char *argv[])
{
	long int min_prec, max_prec, prec, end_flag;
	long int maxsec = 600;

	if(argc <= 2)
	{
		usage(argv[0]);
		return 0;
	}
	else if(argc >= 3)
	{
		min_prec = atol(argv[1]);
		max_prec = atol(argv[2]);
		if((min_prec < 0) || (min_prec > max_prec))
		{
			printf("Invalid precisions! (min = %ld, max = %ld bits)\n", min_prec, max_prec);
			return -1;
		}
		if(argc >= 4)
		{
			maxsec = atol(argv[3]);
			if(maxsec < 0)
			{
				printf("Invalid Maximum seconds! (%ld seconds)\n", maxsec);
				return -1;
			}
		}
	}

	printf("----------------- Benchmark Test using Basic Arithmetics -----------------\n");

//	basebenchloop_by_clock(-2, min_prec, max_prec, 2, maxsec);
//	basebenchloop_by_clock(0, min_prec, max_prec, 2, maxsec);

	basebenchloop(0, min_prec, max_prec, 2, maxsec);
	basebenchloop(1, min_prec, max_prec, 2, maxsec);
#ifdef USE_MPFR
	basebenchloop(2, min_prec, max_prec, 2, maxsec);
	basebenchloop(3, min_prec, max_prec, 2, maxsec);
	basebenchloop(4, min_prec, max_prec, 2, maxsec);
	basebenchloop(5, min_prec, max_prec, 2, maxsec);
	basebenchloop(6, min_prec, max_prec, 2, maxsec);
	basebenchloop(7, min_prec, max_prec, 2, maxsec);
#endif

	printf("---------------------------------- END -----------------------------------\n");
}

