/*----------------------------------------*/
/* Benchmark Test                         */
/*     using Matrix-matrix Multiplication */
/* Copyright (c) 2004- Tomonori Kouya     */
/*                                        */
/* Version 0.0 : Nov. 8, 2004             */
/*----------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bnc.h"

// 256k bytes
//#define CACHESIZE (256 * 1024)

// 512k bytes
//#define CACHESIZE (512 * 1024)

// 1M bytes
//#define CACHESIZE (1 * 1024 * 1024)

// 2M bytes
//#define CACHESIZE (2 * 1024 * 1024)

void usage(const char *commandname)
{
	printf("USAGE-> %s Cachesize(Kilo bytes) [Maximum seconds]\n", commandname);
}

int matmulloop(long int start_dim, long int end_dim, long int dim_step, long int maxsec, long int opt_dim_flag)
{
	long int dim;
	long int i, j, end_flag;
	double iteration, etime, stime;
	DMatrix da, db, dc;

	for(dim = start_dim; dim <= end_dim; dim += dim_step)
	{
		da = init_dmatrix(dim, dim);
		db = init_dmatrix(dim, dim);
		dc = init_dmatrix(dim, dim);

		for(i = 0; i < dim; i++)
		{
			for(j = 0; j < dim; j++)
			{
				set_dmatrix_ij(da, i, j, (i + j + 1) * sqrt(5.0));
				set_dmatrix_ij(db, i, j, (dim - i) * sqrt(3.0));
				set_dmatrix_ij(dc, i, j, sqrt(2.0));
			}
		}

		iteration = 1.0;
		do{
			stime = get_secv();
			for(i = 0; i < iteration; i++) mul_dmatrix(dc, da, db);
			etime = get_secv() - stime;
			iteration *= 2.0;
		} while(etime < 2);

		if(etime > maxsec)
			end_flag = 1;

		etime /= iteration / 2;

		printf("%5d x %5d: %9.5f sec (%8.5f micro-sec/mul)", dim, dim, etime, (etime / dim / dim / dim) * 1000 * 1000);
		if(opt_dim_flag == 1)
			printf(" <-- Optimal dimension %+d", dim - (end_dim + start_dim) / 2);
		printf("\n");

		free_dmatrix(da);
		free_dmatrix(db);
		free_dmatrix(dc);
	}

	return end_flag;
}

int main(int argc, char *argv[])
{
	long int dim, idim, opt_dim, end_flag;
	long int cachesize = 128, maxsec = 600;

	if(argc <= 1)
	{
		usage(argv[0]);
		return 0;
	}
	else if(argc >= 2)
	{
		cachesize = atol(argv[1]);
		cachesize *= 1024;
		if(cachesize < 0)
		{
			printf("Invalid cache memory size! (%ld kilo bytes)\n", cachesize);
			return -1;
		}
		if(argc >= 3)
		{
			maxsec = atol(argv[2]);
			if(maxsec < 0)
			{
				printf("Invalid Maximum seconds! (%ld seconds)\n", maxsec);
				return -1;
			}
		}
	}

	printf("------------------ Benchmark Test using Matrix Multiplication -----------------\n");

//	opt_dim = (long int)floor(sqrt((double)cachesize / 8.0 / 2));
	opt_dim = (long int)floor((-1.0 + sqrt(4 * (double)cachesize / 8.0)) / 2.0);
	printf("Cache size                 : %d Kilo bytes\n", cachesize / 1024);
	printf("Maximum limit of dimension : %d x %d\n\n", opt_dim, opt_dim);

	for(idim = 32; idim <= 16384; idim *= 2)
	{
		end_flag = 0;

		end_flag = matmulloop(idim - 1, idim + 1, 1, maxsec, 0);

		if((idim <= opt_dim) && (opt_dim <= idim * 2))
			end_flag = matmulloop(opt_dim - 5, opt_dim + 5, 1, maxsec, 1);

		if(end_flag == 1)
			break;

	}
	printf("------------------------------------ END --------------------------------------\n");
}

