/************************************************/
/* ex_harmonic.c: Extrapolation harmonic Method           */
/************************************************/
#include <stdio.h>
#include <math.h>
#include "bncode.h"

/* Harmonic sequence */
int dex_harmonic_1step_lo(DVector y, double x0, DVector y0, double h, DMatrix mat, void (* gfunc)(DVector, double), double rtol, double atol)
{
	static long int i, j, k, istep, jstep, flag = 0, dim;
	static long istep_l;
	static double ftmp, ftmp1;

	istep_l = 2;
	dex_harmonic_in_h = h / (double)(istep_l);

	for(istep = 0; istep < dex_harmonic_stage; istep++)
	{

/* First, Using Euler Method */
//		func(dex_harmonic_ytmp[0], x0, y0);
		mul_dmatrix_dvec(dex_harmonic_ytmp[0], mat, y0);
		gfunc(dex_harmonic_ytmp[1], x0);
		add2_dvector(dex_harmonic_ytmp[0], dex_harmonic_ytmp[1]);

		cmul2_dvector(dex_harmonic_ytmp[0], dex_harmonic_in_h);
		add_dvector(dex_harmonic_y[0][istep], y0, dex_harmonic_ytmp[0]);

		subst_dvector(dex_harmonic_ytmp[0], y0);
		subst_dvector(dex_harmonic_ytmp[1], dex_harmonic_y[0][istep]);

/* Next, Using Midpoint Rule */
		for(i = 1; i <= istep_l; i++)
		{
			ftmp = x0 + dex_harmonic_in_h * i;
//			func(dex_harmonic_ytmp[2], ftmp, dex_harmonic_ytmp[1]);
			mul_dmatrix_dvec(dex_harmonic_ytmp[2], mat, dex_harmonic_ytmp[1]);
			gfunc(dex_harmonic_ytmp[3], ftmp);
			add2_dvector(dex_harmonic_ytmp[2], dex_harmonic_ytmp[3]);

			cmul2_dvector(dex_harmonic_ytmp[2], dex_harmonic_in_h);
			for(j = 0; j < y->dim; j++)
				sdvi(dex_harmonic_ytmp[2], j, gdvi(dex_harmonic_ytmp[2], j) * 2);
			add_dvector(dex_harmonic_y[istep][0], dex_harmonic_ytmp[0], dex_harmonic_ytmp[2]);
			if(i == istep_l) /* for smoothing */
				subst_dvector(dex_harmonic_ytmp[2], dex_harmonic_ytmp[0]);

			subst_dvector(dex_harmonic_ytmp[0], dex_harmonic_ytmp[1]);
			subst_dvector(dex_harmonic_ytmp[1], dex_harmonic_y[istep][0]);
		}

/* Smoothing */
		add2_dvector(dex_harmonic_y[istep][0], dex_harmonic_ytmp[2]);
		cmul2_dvector(dex_harmonic_ytmp[0], 2.0);
		add2_dvector(dex_harmonic_y[istep][0], dex_harmonic_ytmp[0]);
		cmul2_dvector(dex_harmonic_y[istep][0], 0.25);

/* Extrapolation */
		for(jstep = 1; jstep <= istep; jstep++)
		{
			flag = 0;
			for(i = 0; i < y->dim; i++)
			{
				sdvi(dex_harmonic_ytmp[2], i, gdvi(dex_harmonic_y[istep][jstep-1], i) - gdvi(dex_harmonic_y[istep-1][jstep-1], i));
				sdvi(dex_harmonic_ytmp[2], i, gdvi(dex_harmonic_ytmp[2], i) / ex_harmonic_den(istep, jstep));
				sdvi(dex_harmonic_y[istep][jstep], i, gdvi(dex_harmonic_y[istep][jstep-1], i) + gdvi(dex_harmonic_ytmp[2], i));

//				if(gdvi(dex_harmonic_y[istep][jstep], i) != gdvi(dex_harmonic_y[istep][jstep-1], i))
//					flag = 1;
			}

			sub_dvector(dex_harmonic_ytmp[1], dex_harmonic_y[istep][jstep], dex_harmonic_y[istep][jstep-1]);
			ftmp  = normi_dvector(dex_harmonic_ytmp[1]);
			ftmp1 = normi_dvector(dex_harmonic_y[istep][jstep]) * rtol + atol;
			if(ftmp <= ftmp1)
				flag = 0;
			else
				flag = 1;
			if(flag == 0)
				goto Return;
		}
	/* 2, 4, 6, 8, 10, ... */
		istep_l += 2;
//		printf("%d(%d) ", istep_l, dex_harmonic_stage);
		dex_harmonic_in_h = h / (double)(istep_l);
	}

Return:
	if(istep == dex_harmonic_stage)
		istep--;
	if(jstep == dex_harmonic_stage)
		jstep--;
	subst_dvector(y, dex_harmonic_y[istep][jstep]);

	if(flag == 0) /* equal */
		return 0;
	else		 /* not equal */
		return 1;
}


/*                          */
/* harmonic Extrapolation Method */
/*                          */
void dex_harmonic_fs_lo(FILE *fp, double x, DVector y, double x0, DVector y0, long int div_num, DMatrix mat, void (* gfunc)(DVector, double), double rtol, double atol, long int stage)
{
	long int steps, i;
	DVector old_y, tmp_y, lf_tmp;
	double tmp[3], new_x, old_x, in_h;
	double start_time, end_time;

	/* time */
	start_time = get_secv();
	
	/* init */
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	if(stage + 1 > MAX_DEX_HARMONIC_STAGE)
	{
		fprintf(stderr, "dex_harmonic_fs: Number of stages exceed.\n");
		return;
	}
	init_dex_harmonic(stage + 1, y->dim);

	/* set */
	old_x = x0;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_h := (x - x0) / div_num */
	if(div_num <= 0)
	{
		fprintf(stderr, "dex_harmonic_fs: Number of division is illegal.\n");
		return;
	}
	in_h = (x - x0) / div_num;

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf("\n");
	}

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* tmp_y := y0 + h * ex_harmonic(x0, y0) */
		dex_harmonic_1step_lo(tmp_y, old_x, old_y, in_h, mat, gfunc, rtol, atol);

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
		if(fp != NULL)
		{
			fprintf(fp, "%25.17e", old_x);
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
				fprintf(fp, "%25.17e", gdvi(old_y, i));
				fprintf(fp, " ");
			}
			fprintf(fp, "\n");
		}

	}

	/* finish! */
	subst_dvector(y, old_y);

	/* time */
	end_time = get_secv();

	printf("OK! dex_harmonic_fs has been just finished.\n");
	printf("Stepsize  : ");
		 printf("%25.17e", in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Computational Time(sec): %f\n", end_time - start_time);
	printf("Integral interval : ");
		 printf("[");
		 printf("%25.17e", x0);
		 printf(", ");
		 printf("%25.17e", old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);

	clear_dex_harmonic();
}

/*                          */
/* harmonic Extrapolation Method */
/*                          */
void dex_harmonic_lo(FILE *fp, double x, DVector y, double x0, DVector y0, double max_h, DMatrix mat, void (* gfunc)(DVector, double), double rtol, double atol, long int stage)
{
	long int steps, i;
	DVector new_y[2], old_y, tmp_y, lf_tmp;
	double tmp[3], new_x, old_x, in_max_h, in_h, min_h;
	double start_time, end_time;

	/* time */
	start_time = get_secv();
	
	/* init */
	new_y[0] = init_dvector(y->dim);
	new_y[1] = init_dvector(y->dim);
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	if(stage + 1 > MAX_DEX_HARMONIC_STAGE)
	{
		fprintf(stderr, "dex_harmonic: Number of stages exceed.\n");
		return;
	}
	init_dex_harmonic(stage + 1, y->dim);

	/* set */
	old_x = x0;
	min_h = max_h;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_max_h = (x - x0) / 2;
	if(max_h < in_max_h)
		in_max_h = max_h;

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf(" stepsize   ");
		printf("\n");
	}

	/* main loop */
	in_h = max_h;
//	in_h = 1.0e-10;
	for(steps = 0; ; steps++)
	{

		/* set default interval */
//		in_h = max_h;
		in_h *= 2;
		if(in_h > max_h)
			in_h = max_h;

		/* check */
		tmp[0] = old_x + in_h;
		if(tmp[0] > x)
			in_h = x - old_x;

		/* calc euler */
		while(1)
		{
			/* tmp_y := y0 + h * ex(x0, y0) */
			if(dex_harmonic_1step_lo(tmp_y, old_x, old_y, in_h, mat, gfunc, rtol, atol) == 0)
				break;

			/* h := h / 2 */
			tmp[2] = in_h / 2;

			/* check in_h */
			tmp[0] = old_x + tmp[2];
			if(tmp[0] <= old_x)
			{
				while(tmp[0] <= old_x)
				{
					in_h *= 2;
					tmp[0] = old_x + in_h;
				}
				dex_harmonic_1step_lo(tmp_y, old_x, old_y, in_h, mat, gfunc, rtol, atol);
			//	fprintf(stderr, "dex_harmonic: inner stepsize is too small.(%15.7e)\n", in_h);
				break;
			}
			else
				in_h = tmp[2];
		}
		if(min_h > in_h)
			min_h = in_h;

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
		if(fp != NULL)
		{
//			ansfunc(tmp_y, old_x);
			fprintf(fp, "%25.17e", old_x);
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
				fprintf(fp, "%25.17e", gdvi(old_y, i));
//				fprintf(fp, "%10.7e", fabs(gdvi(old_y, i) - gdvi(tmp_y, i))/fabs(gdvi(tmp_y, i)));
				fprintf(fp, " ");
			}
			fprintf(fp, "%15.7e ", in_h);
			fprintf(fp, "\n");
		}

		if(old_x >= x)
			break;

	}

	/* finish! */
	subst_dvector(y, old_y);

	/* time */
	end_time = get_secv();

	printf("OK! dex_harmonic has been just finished.\n");
	printf("Miharmonicum/Maxmum stepsize  : ");
		 printf("%15.7e", min_h);
		 printf(" / ");
		 printf("%15.7e", max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Computational Time(sec): %f\n", end_time - start_time);
	printf("Integral interval : ");
		 printf("[");
		 printf("%15.7e", x0);
		 printf(", ");
		 printf("%15.7e", old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_dvector(new_y[0]);
	free_dvector(new_y[1]);
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);

	clear_dex_harmonic();

}

#ifdef USE_GMP

/*                          */
/* harmonic Extrapolation Method */
/*                          */
int mpf_ex_harmonic_1step_lo(MPFVector y, mpf_t x0, MPFVector y0, mpf_t h, MPFMatrix mat, void (* gfunc)(MPFVector, mpf_t), mpf_t rtol, mpf_t atol)
{
	long int i, j, k, istep, jstep, flag = 0, dim;
	long istep_l;

	istep_l = 2;
	mpf_div_ui(ex_harmonic_in_h, h, (unsigned long)istep_l);

	for(istep = 0; istep < ex_harmonic_stage; istep++)
	{

/* First, Using Euler Method */
//		func(ex_harmonic_ytmp[0], x0, y0);
		mul_mpfmatrix_mpfvec(ex_harmonic_ytmp[0], mat, y0);
		gfunc(ex_harmonic_ytmp[1], x0);
		add2_mpfvector(ex_harmonic_ytmp[0], ex_harmonic_ytmp[1]);

		cmul2_mpfvector(ex_harmonic_ytmp[0], ex_harmonic_in_h);
		add_mpfvector(ex_harmonic_y[0][istep], y0, ex_harmonic_ytmp[0]);

		subst_mpfvector(ex_harmonic_ytmp[0], y0);
		subst_mpfvector(ex_harmonic_ytmp[1], ex_harmonic_y[0][istep]);

/* Next, Using Midpoint Rule */
		for(i = 1; i <= istep_l; i++)
		{
			mpf_mul_ui(ex_harmonic_mpf_tmp, ex_harmonic_in_h, (unsigned long)i);
			mpf_add(ex_harmonic_mpf_tmp, x0, ex_harmonic_mpf_tmp);
//			func(ex_harmonic_ytmp[2], ex_harmonic_mpf_tmp, ex_harmonic_ytmp[1]);
			mul_mpfmatrix_mpfvec(ex_harmonic_ytmp[2], mat, ex_harmonic_ytmp[1]);
			gfunc(ex_harmonic_ytmp[3], ex_harmonic_mpf_tmp);
			add2_mpfvector(ex_harmonic_ytmp[2], ex_harmonic_ytmp[3]);

			cmul2_mpfvector(ex_harmonic_ytmp[2], ex_harmonic_in_h);
			for(j = 0; j < y->dim; j++)
				mpf_mul_ui(
					get_mpfvector_i(ex_harmonic_ytmp[2], j),
					get_mpfvector_i(ex_harmonic_ytmp[2], j),
					(unsigned long)2
				);
			add_mpfvector(ex_harmonic_y[istep][0], ex_harmonic_ytmp[0], ex_harmonic_ytmp[2]);
			if(i == istep_l) /* for smoothing */
				subst_mpfvector(ex_harmonic_ytmp[2], ex_harmonic_ytmp[0]);

			subst_mpfvector(ex_harmonic_ytmp[0], ex_harmonic_ytmp[1]);
			subst_mpfvector(ex_harmonic_ytmp[1], ex_harmonic_y[istep][0]);
		}
/* Smoothing */
		add2_mpfvector(ex_harmonic_y[istep][0], ex_harmonic_ytmp[2]);
		mpf_set_ui(ex_harmonic_mpf_tmp, 2UL);
		cmul2_mpfvector(ex_harmonic_ytmp[0], ex_harmonic_mpf_tmp);
		add2_mpfvector(ex_harmonic_y[istep][0], ex_harmonic_ytmp[0]);
		mpf_set_ui(ex_harmonic_mpf_tmp, 1UL);
		mpf_div_ui(ex_harmonic_mpf_tmp, ex_harmonic_mpf_tmp, 4UL);
		cmul2_mpfvector(ex_harmonic_y[istep][0], ex_harmonic_mpf_tmp);

/* Extrapolation */
		for(jstep = 1; jstep <= istep; jstep++)
		{
			flag = 0;
			mpf_ex_harmonic_den(ex_harmonic_mpf_tmp, istep, jstep);
			for(i = 0; i < y->dim; i++)
			{
				mpf_sub(
					get_mpfvector_i(ex_harmonic_ytmp[2], i),
					get_mpfvector_i(ex_harmonic_y[istep][jstep-1], i),
					get_mpfvector_i(ex_harmonic_y[istep-1][jstep-1], i)
				);
				mpf_div(
					get_mpfvector_i(ex_harmonic_ytmp[2], i),
				 	get_mpfvector_i(ex_harmonic_ytmp[2], i),
				 	ex_harmonic_mpf_tmp
				);
				mpf_add(
					get_mpfvector_i(ex_harmonic_y[istep][jstep], i),
					get_mpfvector_i(ex_harmonic_y[istep][jstep-1], i),
					get_mpfvector_i(ex_harmonic_ytmp[2], i)
				);
			
//				if(mpf_cmp(get_mpfvector_i(ex_harmonic_y[istep][jstep], i), get_mpfvector_i(ex_harmonic_y[istep][jstep-1], i) ) != 0)
//					flag = 1;
			}

			/* (ex_harmonic_ytmp[0], 0) := || ex_harmonic_y[istep][jstep] - ex_harmonic_y[istep][jstep-1] || * rtol + atol */
			sub_mpfvector(ex_harmonic_ytmp[1], ex_harmonic_y[istep][jstep], ex_harmonic_y[istep][jstep-1]);
			normi_mpfvector(get_mpfvector_i(ex_harmonic_ytmp[0], 0), ex_harmonic_ytmp[1]);
	
			/* ex_harmonic_tmp := || ex_harmonic_y[istep][jstep] || */
			normi_mpfvector(ex_harmonic_mpf_tmp, ex_harmonic_y[istep][jstep]);
			mpf_mul(ex_harmonic_mpf_tmp, ex_harmonic_mpf_tmp, rtol);
			mpf_add(ex_harmonic_mpf_tmp, ex_harmonic_mpf_tmp, atol);

			/* || ex_harmonic_y[istep][jstep] - ex_harmonic_y[istep][jstep-1]  || <= || ex_harmonic_y[istep][jstep] || * rtol + atol */
			if(mpf_cmp(gmpfvi(ex_harmonic_ytmp[0], 0), ex_harmonic_mpf_tmp) <= 0)
				flag = 0;
			else
				flag = 1;
		//	print_mpfvector(ex_y[istep][jstep]);
			if(flag == 0)
			{
//				fflush(stdout); printf("return! %ld\n", istep);
				goto Return;
			}
		}
		istep_l += 2;
//		printf("%d(%d) ", istep_l, ex_harmonic_stage);
		mpf_div_ui(ex_harmonic_in_h, h, istep_l);
	}

Return:
	if(istep == ex_harmonic_stage)
		istep--;
	if(jstep == ex_harmonic_stage)
		jstep--;
	subst_mpfvector(y, ex_harmonic_y[istep][jstep]);

	if(flag == 0) /* equal */
		return 0;
	else		 /* not equal */
		return 1;
}

/*                          */
/* harmonic Extrapolation Method */
/* Stepsize fixed           */
void mpf_ex_harmonic_fs_lo(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, long int div_num, MPFMatrix mat, void (* gfunc)(MPFVector, mpf_t), mpf_t rtol, mpf_t atol, long int stage)
{
	long int steps, i;
	MPFVector old_y, tmp_y, lf_tmp;
	mpf_t tmp[3], new_x, old_x, in_h;
	double start_time, end_time;

	/* time */
	start_time = get_secv();

	/* init */
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);
	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));

	if(stage + 1 > MAX_EX_HARMONIC_STAGE)
	{
		fprintf(stderr, "mpf_ex_harmonic_fs: Number of stages exceed.\n");
		return;
	}
	init_mpf_ex_harmonic(stage + 1, y->dim, y->prec);

	/* set */
	mpf_set(old_x, x0);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_h := (x - x0) / div_num */
	mpf_sub(in_h, x, x0);
	if(div_num <= 0)
	{
		fprintf(stderr, "mpf_ex_harmonic_fs: A number of division is illegal.\n");
		return;
	}
	mpf_div_ui(in_h, in_h, div_num);

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf("\n");
	}

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* calc euler */
		/* tmp_y := y0 + h * euler(x0, y0) */
		mpf_ex_harmonic_1step_lo(tmp_y, old_x, old_y, in_h, mat, gfunc, rtol, atol);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
		if(fp != NULL)
		{
//			mpf_out_str(fp, 10, 0, old_x);
			fprintf(fp, "%25.17e", mpf2double(old_x));
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
//				mpf_out_str(fp, 10, 0, gmpfvi(old_y, i));
				fprintf(fp, "%25.17e", mpf_get_d(gmpfvi(old_y, i)));
				fprintf(fp, " ");
			}
			fprintf(fp, "\n");
		}

	}

	/* finish! */
	subst_mpfvector(y, old_y);

	/* time */
	end_time = get_secv();

	printf("OK! mpf_ex_harmonic_fs has been just finished.\n");
	printf("Stepsize  : ");
		 mpf_out_str(stdout, 10, 0, in_h);
//		 printf(", Number of division: %ld ", div_num);
		 printf("\n");
	printf("Number of steps        : %ld\n", steps);
	printf("Computational Time(sec): %f\n", end_time - start_time);
	printf("Integral interval : ");
		 printf("[");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);

	clear_mpf_ex_harmonic();

}

/*                          */
/* harmonic Extrapolation Method */
/*                          */
void mpf_ex_harmonic_lo(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, mpf_t max_h, MPFMatrix mat, void (* gfunc)(MPFVector, mpf_t), mpf_t rtol, mpf_t atol, long int stage)
{
	long int steps, i;
	MPFVector new_y[2], old_y, tmp_y, lf_tmp;
	mpf_t tmp[3], new_x, old_x, in_max_h, in_h, min_h;
	double start_time, end_time;

	/* time */
	start_time = get_secv();
	
	/* init */
	new_y[0] = init2_mpfvector(y->dim, y->prec);
	new_y[1] = init2_mpfvector(y->dim, y->prec);
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);
	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));
	mpf_init2(in_max_h, mpf_get_prec(x));
	mpf_init2(min_h, mpf_get_prec(x));

	if(stage + 1 > MAX_EX_HARMONIC_STAGE)
	{
		fprintf(stderr, "mpf_ex_harmonic: Number of stages exceed.\n");
		return;
	}
	init_mpf_ex_harmonic(stage + 1, y->dim, y->prec);

	/* set */
	mpf_set(old_x, x0);
	mpf_set(min_h, max_h);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	mpf_sub(in_max_h, x, x0);
	mpf_div_ui(in_max_h, in_max_h, 2UL);
	if(mpf_cmp(max_h, in_max_h) < 0)
		mpf_set(in_max_h, max_h);

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf(" stepsize   ");
		printf("\n");
	}

	/* main loop */
	mpf_set(in_h, max_h);
	for(steps = 0; ; steps++)
	{

		/* set default interval */
//		mpf_set(in_h, max_h);
		mpf_mul_ui(in_h, in_h, 2UL);

		/* check */
		mpf_add(tmp[0], old_x, in_h);
		if(mpf_cmp(tmp[0], x) > 0)
			mpf_sub(in_h, x, old_x);

		/* calc euler */
		while(1)
		{
			/* tmp_y := y0 + h * euler(x0, y0) */
			if(mpf_ex_harmonic_1step_lo(tmp_y, old_x, old_y, in_h, mat, gfunc, rtol, atol) == 0)
				break;

			/* h := h / 2 */
			mpf_div_ui(tmp[2], in_h, 2UL);

			/* check in_h */
			mpf_add(tmp[0], old_x, tmp[2]);
			if(mpf_cmp(tmp[0], old_x) <= 0)
			{
				while(mpf_cmp(tmp[0], old_x) <= 0)
				{
					mpf_mul_ui(in_h, in_h, 2UL);
					mpf_add(tmp[0], old_x, in_h);
				}
				mpf_ex_harmonic_1step_lo(tmp_y, old_x, old_y, in_h, mat, gfunc, rtol, atol);
//				fprintf(stderr, "mpf_ex_harmonic: inner stepsize is too small.\n");
				break;
			}
			else
				mpf_set(in_h, tmp[2]);
		}

		if(mpf_cmp(min_h, in_h) > 0)
			mpf_set(min_h, in_h);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
		if(fp != NULL)
//		if((fp != NULL) && ((steps % itimes_iv) == 0))
		{
//			mpf_out_str(fp, 10, 0, old_x);
			fprintf(fp, "%25.17e", mpf2double(old_x));
//			ansfunc(tmp_y, old_x);
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
				mpf_out_str(fp, 10, 0, gmpfvi(old_y, i));
//				fprintf(fp, "%25.17e", mpf_get_d(gmpfvi(old_y, i)));
//				mpf_sub(tmp[0], gmpfvi(old_y, i), gmpfvi(tmp_y, i));
//				mpf_div(tmp[0], tmp[0], gmpfvi(tmp_y, i));
//				mpf_abs(tmp[0], tmp[0]);
//				mpf_out_str(fp, 10, 7, tmp[0]);
				fprintf(fp, " ");
			}
			fprintf(fp, "%15.7e ", mpf2double(in_h));
			fprintf(fp, "\n");
		}

		if(mpf_cmp(old_x, x) >= 0)
			break;

	}

	/* finish! */
	subst_mpfvector(y, old_y);

	printf("OK! mpf_ex_harmonic has been just finished.\n");
	printf("Miharmonicum/Maxmum stepsize  : ");
		 mpf_out_str(stdout, 10, 0, min_h);
		 printf(" / ");
		 mpf_out_str(stdout, 10, 0, max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Computational Time(sec): %f\n", end_time - start_time);
	printf("Integral interval : ");
		 printf("[");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_mpfvector(new_y[0]);
	free_mpfvector(new_y[1]);
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);
	mpf_clear(in_max_h);
	mpf_clear(min_h);

	clear_mpf_ex_harmonic();

}
#endif
