/************************************************/
/* Test program for ex_*.c                      */
/* Copyleft 2004, Tomonori Kouya                */
/************************************************/
#include <stdio.h>
#include <math.h>

//#define USE_GMP
#include "bncode.h"

#define DALPHA 0.99999999
//#define MPFALPHA_U 99999999
//#define MPFALPHA_D 100000000
#define MPFALPHA_U 99999999
#define MPFALPHA_D 100000000
//#define DALPHA 0.9

/* double */
void ddf(DVector y, double x0, DVector y0)
{
	/* y(0)' = y(1) */
	/* y(1)' = -alpha * y(0)^2 * sin(x0) + 2 * alpha * y(0) * y(1) * cos(x0) */
	set_dvector_i(y, 0, gdvi(y0, 1));
	set_dvector_i(y, 1, -DALPHA * gdvi(y0, 0) *gdvi(y0, 0) * sin(x0) + 2 * DALPHA * gdvi(y0, 0) * gdvi(y0, 1) * cos(x0));

	return;
}

/* y(0) = 1/(1-alpha * sin(x)) */
/* y(1) = alpha * cos(x) / (1-alpha * sin(x))^2 */
void dans(DVector y, double x)
{
	long int i;
	double base;

	base = 1.0 - DALPHA * sin(x);

	set_dvector_i(y, 0, 1.0 / base);
	set_dvector_i(y, 1, DALPHA * cos(x) / (base * base));

	return;
}

void initial_dvalue(double *lf_initx, double *lf_maxx, DVector lf_inity)
{
	long int i;

	/* x0 := 0 */
	/* y := [1, ..., 1]^T */
	*lf_initx = 0.0;
	*lf_maxx = 37.0;

	set_dvector_i(lf_inity, 0, 1.0);
	set_dvector_i(lf_inity, 1, DALPHA);
}

#ifdef USE_GMP
/* mpf_t */
void df(MPFVector y, mpf_t x0, MPFVector y0)
{
	static int init_flag = 1;
	mpf_t tmp, tmp1;

//	if(init_flag == 1)
	{
		mpf_init2(tmp, y->prec);
		mpf_init2(tmp1, y->prec);
		init_flag = 0;
	}
	/* y' = y */
	/* y(0)' = y(1) */
	/* y(1)' = -alpha * y(0)^2 * sin(x0) + 2 * alpha * y(0) * y(1) * cos(x0) */
	set_mpfvector_i(y, 0, gmpfvi(y0, 1));

	mpf_sin(tmp, x0);
	mpf_mul_ui(tmp, tmp, MPFALPHA_U);
	mpf_div_ui(tmp, tmp, MPFALPHA_D);
	mpf_mul(tmp, tmp, gmpfvi(y0, 0));
	mpf_mul(tmp, tmp, gmpfvi(y0, 0));
	mpf_neg(tmp, tmp);

	mpf_cos(tmp1, x0);
	mpf_mul_ui(tmp1, tmp1, 2UL);
	mpf_mul_ui(tmp1, tmp1, MPFALPHA_U);
	mpf_div_ui(tmp1, tmp1, MPFALPHA_D);
	mpf_mul(tmp1, tmp1, gmpfvi(y0, 0));
	mpf_mul(tmp1, tmp1, gmpfvi(y0, 1));

	mpf_add(tmp, tmp, tmp1);
	set_mpfvector_i(y, 1, tmp);

	mpf_clear(tmp);
	mpf_clear(tmp1);

	return;
}

/* y(0) = 1/(1-alpha * sin(x)) */
/* y(1) = alpha * cos(x) / (1-alpha * sin(x))^2 */
void ans(MPFVector y, mpf_t x)
{
	long int i;
	mpf_t tmp, tmp1;

	mpf_init2(tmp, y->prec);
	mpf_init2(tmp1, y->prec);

	mpf_sin(tmp, x);
	mpf_mul_ui(tmp, tmp, MPFALPHA_U);
	mpf_div_ui(tmp, tmp, MPFALPHA_D);
	mpf_ui_sub(tmp, 1UL, tmp);
	mpf_ui_div(tmp, 1UL, tmp);

	set_mpfvector_i(y, 0, tmp);

	mpf_mul(tmp, tmp, tmp);
	mpf_mul_ui(tmp, tmp, MPFALPHA_U);
	mpf_div_ui(tmp, tmp, MPFALPHA_D);
	mpf_cos(tmp1, x);
	mpf_mul(tmp, tmp, tmp1);
	
	set_mpfvector_i(y, 1, tmp);

	mpf_clear(tmp);
	mpf_clear(tmp1);
	return;
}


void initial_value(mpf_t lf_initx, mpf_t lf_maxx, MPFVector lf_inity)
{
	long int i;
	mpf_t tmp;

	mpf_init2(tmp, lf_inity->prec);

	/* x0 := 0 */
	/* y := [1, ..., 1]^T */
	mpf_set_ui(lf_initx, 0UL);
	mpf_set_ui(lf_maxx, 37UL);

	mpf_set_ui(tmp, 1UL);
	set_mpfvector_i(lf_inity, 0, tmp);
	mpf_set_ui(tmp, MPFALPHA_U);
	mpf_div_ui(tmp, tmp, MPFALPHA_D);
	set_mpfvector_i(lf_inity, 1, tmp);

	mpf_clear(tmp);
}
#endif

#define DIM 2

main()
{
	DVector dy, init_dy, lf_dtmp;
	long int div_num;
	double dx, init_dx, dstepsize, dabs_tol, drel_tol;
	double stime, etime;
#ifdef USE_GMP
	MPFVector y, init_y, lf_tmp;
	mpf_t x, init_x, stepsize, abs_tol, rel_tol;
#endif


	dy = init_dvector(DIM);
	lf_dtmp = init_dvector(DIM);
	init_dy = init_dvector(DIM);
	initial_dvalue(&init_dx, &dx, init_dy);

	dstepsize = 1.0;
	dabs_tol = 0.0;
	drel_tol = 1.0e-35;
//	drel_tol = 1.0e-20;
//	drel_tol = 1.0e-25;
//	drel_tol = 0.0;

//	goto HARMONIC;

	/* Extrapolation */
	printf("-- ex_nim(double) -- \n");

	initial_dvalue(&init_dx, &dx, init_dy);

//	dex_nim(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 4, dans);
//	dex_nim(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 8, dans);
//	dex_nim(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 16, dans);

	printf("-- ex_nim(double) ended -- \n");

//	goto END;

HARMONIC:

	printf("-- ex_harmonic(double) -- \n");

	initial_dvalue(&init_dx, &dx, init_dy);
//	dex_harmonic(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 4, dans);
//	dex_harmonic(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 8, dans);
//	dex_harmonic(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 16, dans);

	printf("-- ex_harmonic(double) ended -- \n");
//	goto END;

#ifdef USE_GMP
//	set_bnc_default_prec(128);
	set_bnc_default_prec_decimal(50);

	/* x := 1 */
	mpf_init(init_x);
	mpf_init(x);

	y = init_mpfvector(DIM);
	lf_tmp = init_mpfvector(DIM);
	init_y = init_mpfvector(DIM);
	initial_value(init_x, x, init_y);

	mpf_init_set_d(stepsize, dstepsize);
//	printf("max.stepsize: "); mpf_out_str(stdout, 10, 0, stepsize); printf("\n");
	mpf_init_set_d(abs_tol, dabs_tol);
//	printf("Abs.tol: "); mpf_out_str(stdout, 10, 0, abs_tol); printf("\n");
	mpf_init_set_d(rel_tol, drel_tol);
//	printf("Rel.tol: "); mpf_out_str(stdout, 10, 0, rel_tol); printf("\n");
//	mpf_init_set_d(rel_tol, 1.0e-20);

/* 4stages, 128bits */
/*
-- ex_nim ended --

real    2m57.379s
user    2m53.577s
sys     0m0.249s
*/
	printf("-- ex_nim -- \n");

//	mpf_ex_nim(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 4, ans, 1);
	mpf_ex_nim(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 8, ans, 1);
//	mpf_ex_nim(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 16, ans, 1);

	printf("-- ex_nim ended -- \n");

/*-- ex_harmonic ended --

real    2m35.402s
user    2m29.843s
sys     0m0.531s
*/
	printf("-- ex_harmonic -- \n");

//	mpf_ex_harmonic(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 4, ans, 1);
//	mpf_ex_harmonic(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 8, ans, 1);
//	mpf_ex_harmonic(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 16, ans, 1);
#endif
	printf("-- ex_harmonic ended -- \n");

END:

}
