/************************************************/
/* Test program for ex_*.c                      */
/* Copyleft 2004, Tomonori Kouya                */
/************************************************/
#include <stdio.h>
#include <math.h>

//#define USE_GMP
#include "bncode.h"

/* double */
void ddf(DVector y, double x0, DVector y0)
{
	long int i;

	/* y_i' = i*y_i */
	for(i = 0; i < y->dim; i++)
		set_dvector_i(y, i, (i + 1) * get_dvector_i(y0, i));

	return;
}

/* y = [exp(x), ...., n*exp(x)]^T */
void dans(DVector y, double x)
{
	long int i;

	for(i = 0; i < y->dim; i++)
		set_dvector_i(y, i, exp((i + 1) * x));

	return;
}

#ifdef USE_GMP
/* mpf_t */
void df(MPFVector y, mpf_t x0, MPFVector y0)
{
	long int i;
	mpf_t tmp;

	mpf_init2(tmp, y->prec);

	/* y' = y */
	for(i = 0; i < y->dim; i++)
	{
		mpf_mul_ui(tmp, get_mpfvector_i(y0, i), i + 1);
		set_mpfvector_i(y, i,tmp);
	}

	mpf_clear(tmp);

	return;
}

/* y = [exp(x), ...., exp(x)]^T */
void ans(MPFVector y, mpf_t x)
{
	long int i;
	mpf_t tmp;

	mpf_init2(tmp, y->prec);
	for(i = 0; i < y->dim; i++)
	{
		mpf_mul_ui(tmp, x, i + 1);
		mpf_exp(tmp, tmp);
		set_mpfvector_i(y, i, tmp);
	}
	mpf_clear(tmp);
	return;
}


void initial_value(mpf_t lf_initx, mpf_t lf_maxx, MPFVector lf_inity)
{
	long int i;

	/* x0 := 0 */
	/* y := [1, ..., 1]^T */
	mpf_init_set_str(lf_initx, "0.0", 10);
	mpf_init_set_str(lf_maxx, "1.0", 10);
	for(i = 0; i < lf_inity->dim; i++)
		set_mpfvector_i_str(lf_inity, i, "1.0", 10);
}
#endif

void initial_dvalue(double *lf_initx, double *lf_maxx, DVector lf_inity)
{
	long int i;

	/* x0 := 0 */
	/* y := [1, ..., 1]^T */
	*lf_initx = 0.0;
	*lf_maxx = 1.0;
	for(i = 0; i < lf_inity->dim; i++)
		set_dvector_i(lf_inity, i, 1.0);
}

#define DIM 100

main()
{
	DVector dy, init_dy, lf_dtmp;
	long int div_num;
	double dx, init_dx, dstepsize, dabs_tol, drel_tol;
	double stime, etime;
#ifdef USE_GMP
	MPFVector y, init_y, lf_tmp;
	mpf_t x, init_x, stepsize, abs_tol, rel_tol;

	set_bnc_default_prec_decimal(50);

	/* x := 1 */
	mpf_init(init_x);
	mpf_init(x);

	y = init_mpfvector(DIM);
	lf_tmp = init_mpfvector(DIM);
	init_y = init_mpfvector(DIM);
	initial_value(init_x, x, init_y);
#endif

	dy = init_dvector(DIM);
	lf_dtmp = init_dvector(DIM);
	init_dy = init_dvector(DIM);
	initial_dvalue(&init_dx, &dx, init_dy);

	dstepsize = 0.5;
	dabs_tol = 0.0;
//	drel_tol = 1.0e-15;
	drel_tol = 0.0;

#ifdef USE_GMP
	mpf_init_set_d(stepsize, dstepsize);
	mpf_init_set_d(abs_tol, dabs_tol);
	mpf_init_set_d(rel_tol, drel_tol);
#endif

	/* Extrapolation */
	printf("-- ex_nim -- \n");
	for(div_num = 2; div_num <= 128; div_num *= 2)
	{
		initial_dvalue(&init_dx, &dx, init_dy);
		stime = get_secv();
		dex_nim_fs(NULL, dx, dy, init_dx, init_dy, div_num, ddf, drel_tol, dabs_tol, 4);
		etime = get_secv();
		printf("time(sec): %f\n", etime - stime);
		dans(init_dy, dx);
		dx = normi_dvector(init_dy);
		sub_dvector(init_dy, init_dy, dy);
		init_dx = normi_dvector(init_dy);
		printf("error, rerror(infty norm): %10.3e, %10.3e\n\n", init_dx, init_dx/dx);
#ifdef USE_GMP
		initial_value(init_x, x, init_y);
		stime = get_secv();
		mpf_ex_nim_fs(NULL, x, y, init_x, init_y, div_num, df, rel_tol, abs_tol, 8);
		etime = get_secv();
		printf("time(sec): %f\n", etime - stime);
		ans(init_y, x);
		normi_mpfvector(x, init_y);
		sub_mpfvector(init_y, init_y, y);
		normi_mpfvector(init_x, init_y);
		printf("error, rerror(infty norm): %10.3e, %10.3e\n\n", mpf_get_d(init_x), mpf_get_d(init_x)/mpf_get_d(x));
#endif
	}
//	dex_nim(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 4, dans);
//	dex_nim(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 8, dans);
//	dex_nim(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 16, dans);
#ifdef USE_GMP
//	mpf_ex_nim(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 4);
//	mpf_ex_nim(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 8);
//	mpf_ex_nim(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 16);
#endif
	printf("-- ex_nim ended -- \n");

	/* Extrapolation */
	printf("-- ex_harmonic -- \n");
	for(div_num = 2; div_num <= 128; div_num *= 2)
	{
		initial_dvalue(&init_dx, &dx, init_dy);
		stime = get_secv();
		dex_harmonic_fs(NULL, dx, dy, init_dx, init_dy, div_num, ddf, drel_tol, dabs_tol, 4);
		etime = get_secv();
		printf("time(sec): %f\n", etime - stime);
		dans(init_dy, dx);
		dx = normi_dvector(init_dy);
		sub_dvector(init_dy, init_dy, dy);
		init_dx = normi_dvector(init_dy);
		printf("error, rerror(infty norm): %10.3e, %10.3e\n\n", init_dx, init_dx/dx);
#ifdef USE_GMP
		initial_value(init_x, x, init_y);
		stime = get_secv();
		mpf_ex_harmonic_fs(NULL, x, y, init_x, init_y, div_num, df, rel_tol, abs_tol, 8);
		etime = get_secv();
		printf("time(sec): %f\n", etime - stime);
		ans(init_y, x);
		normi_mpfvector(x, init_y);
		sub_mpfvector(init_y, init_y, y);
		normi_mpfvector(init_x, init_y);
		printf("error, rerror(infty norm): %10.3e, %10.3e\n\n", mpf_get_d(init_x), mpf_get_d(init_x)/mpf_get_d(x));
#endif
	}
//	dex_harmonic(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 4, dans);
//	dex_harmonic(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 8, dans);
//	dex_harmonic(stdout, dx, dy, init_dx, init_dy, dstepsize, ddf, drel_tol, dabs_tol, 16, dans);
#ifdef USE_GMP
//	mpf_ex_harmonic(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 4);
//	mpf_ex_harmonic(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 8);
//		stime = get_secv();
//	mpf_ex_harmonic(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 16);
//	mpf_ex_harmonic(stdout, x, y, init_x, init_y, stepsize, df, rel_tol, abs_tol, 16);
//		etime = get_secv();
//		printf("time(sec): %f\n", etime - stime);
#endif
	printf("-- ex_harmonic ended -- \n");


}
