/*****************************************************************************/
/* Extented gexpr.c by MPFR                                                  */
/* Original gepr.c from http://swox.com/gmp/gexpr.c                          */
/* ************************** CAUTION! ***************************************/
/* GNU MP and MPFR are based on LGPL(Lesser GNU Public License),             */
/* also BNCpack and MPIBNCpack. BUT the "mpfr_gexpr.c"'s copyright is GPL    */
/* because of GPLed original "gexpr.c."                                      */
/*                                                                           */
/* Expression evaluation using plain floating-point arithmetic.              */
/* Copyright (C) 1999, 2000, 2001, 2002, 2003 Free Software Foundation, Inc. */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2, or (at your option) any later */
/* version. This program is distributed in the hope that it will be useful,  */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; see the file COPYING.  If not, write to the Free       */
/* Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,  */
/* USA.                                                                      */
/*****************************************************************************/
#include <string.h>
#include <stdio.h>
#include <setjmp.h>
#include <math.h>

#ifdef USE_GMP
#include "gmp.h"
#include "mpfr.h"

jmp_buf top;

static char *bp, *expr_start_p, *expr_end_p;
static int ch;
static int previous_result_valid_flag;

mp_rnd_t default_rmode = GMP_RNDN; // round to nearest
mpfr_t previous_result;

void mpfr_term(mpfr_t ret), mpfr_expo(mpfr_t ret), mpfr_factor(mpfr_t ret), mpfr_number(mpfr_t ret);


#define next skip ()

void
skip ()
{
  do
    ch = *bp++;
  while (ch == ' ' || ch == '\n');
}

void
new_error (char *funcname)
{
  fprintf (stderr, "Syntax error in %s\n", funcname);
  longjmp (top, 1);
}
void
error ()
{
  fprintf (stderr, " ^ syntax error\n");
  longjmp (top, 1);
}

void mpfr_expr(mpfr_t ret)
{
  mpfr_t tmp;

//  printf("mpfr_expr!\n"); fflush(stdout);

  mpfr_init2(tmp, mpfr_get_prec(ret));

  if (ch == '+')
    {
      next;
      mpfr_term(ret);
    }
  else if (ch == '-')
    {
      next;
      mpfr_term(ret);
      mpfr_neg(ret, ret, default_rmode);
    }
  else
    mpfr_term(ret);

  while (ch == '+' || ch == '-')
    { char op;
      op = ch;
      next;
      if (op == '-')
      {
        mpfr_term(tmp);
	mpfr_sub(ret, ret, tmp, default_rmode);
      }
      else
      {
        mpfr_term(tmp);
	mpfr_add(ret, ret, tmp, default_rmode);
      }
    }

    mpfr_clear(tmp);
 //printf("mpfr_expr ended!\n"); fflush(stdout);
}

void mpfr_term(mpfr_t ret)
{
  mpfr_t tmp;

  //printf("mpfr_term!\n"); fflush(stdout);

  mpfr_init2(tmp, mpfr_get_prec(ret));

  mpfr_expo(ret);
  for (;;)
    switch (ch)
      {
      case '*':
	next;
	mpfr_expo(tmp);
	mpfr_mul(ret, ret, tmp, default_rmode);
	break;
      case '/':
	next;
	mpfr_expo(tmp);
	mpfr_div(ret, ret, tmp, default_rmode);
	break;
/*      case '%':
	next;
	t = fmod (t, expo ());
	break;
*/
      default:
  	 mpfr_clear(tmp);
//  printf("mpfr_term ended!\n"); fflush(stdout);
	return;
      }

}

// ret^expo() = exp(expo() * log(ret))
void mpfr_expo(mpfr_t ret)
{
  mpfr_t tmp, tmp1;

 // printf("mpfr_expo!\n"); fflush(stdout);

  mpfr_init2(tmp,  mpfr_get_prec(ret));
  mpfr_init2(tmp1, mpfr_get_prec(ret));

  mpfr_factor(ret);
  if (ch == '^')
    {
      next;
      //e = pow (e, expo ());
      mpfr_expo(tmp);
      mpfr_log(tmp1, ret, default_rmode); // log_e (ret)
      mpfr_mul(ret, tmp, tmp1, default_rmode);
      mpfr_exp(ret, ret, default_rmode); // e^(expo() * log_e(ret)))
    }

   mpfr_clear(tmp);
   mpfr_clear(tmp1);
  //printf("mpfr_expo ended!\n"); fflush(stdout);
}

struct functions
{
  char *spelling;
  //int (* evalfn) (mpfr_t, mpfr_t, mp_rnd_t);
  int (* evalfn)(mpfr_ptr, mpfr_srcptr, mp_rnd_t);
};

/* double
log2 (x)
     double x;
{
  return log (x) * 1.442695040888963;
}
*/

int mpfr_floor_dummy(mpfr_ptr rop, mpfr_srcptr op, mp_rnd_t rmode)
{
	mpfr_floor(rop, op);
}

int mpfr_ceil_dummy(mpfr_ptr rop, mpfr_srcptr op, mp_rnd_t rmode)
{
	mpfr_ceil(rop, op);
}

struct functions fns[] =
{
//  {"log2", log2},
  {"log2", mpfr_log2},
  {"log10", mpfr_log10},
  {"log", mpfr_log},
  {"exp", mpfr_exp},
  {"sqrt", mpfr_sqrt},
  {"floor", mpfr_floor_dummy},
  {"ceil", mpfr_ceil_dummy},
  {"sin", mpfr_sin},
  {"cos", mpfr_cos},
  {"tan", mpfr_tan},
  {"asin", mpfr_asin},
  {"acos", mpfr_acos},
  {"atan", mpfr_atan},
  {"sinh", mpfr_sinh},
  {"cosh", mpfr_cosh},
  {"tanh", mpfr_tanh},
  {"asinh", mpfr_asinh},
  {"acosh", mpfr_acosh},
  {"atanh", mpfr_atanh},
#ifdef mpfr_cmp_abs // newly defined in mpfr-2.0.2
  {"expm1", mpfr_expm1},
  {"gamma", mpfr_gamma},
  {"zeta", mpfr_zeta},
  {"erf", mpfr_erf},
#endif
  {0, 0}
};


void mpfr_factor(mpfr_t ret)
{
  mpfr_t tmp;
  int i;

  //printf("mpfr_factor!\n"); fflush(stdout);

  mpfr_init2(tmp, mpfr_get_prec(ret));

  for (i = 0; fns[i].spelling != 0; i++)
    {
      char *spelling = fns[i].spelling;
      int len = strlen (spelling);
      if (strncmp (spelling, bp - 1, len) == 0 && ! isalnum (bp[-1 + len]))
	{
	  bp += len - 1;
	  next;
	  if (ch != '(')
	    //error ();
	    new_error ("mpfr_factor 1");
	  next;
	  mpfr_expr(tmp);
	  if (ch != ')')
	  //  error ();
	    new_error ("mpfr_factor 2");
	  next;
	  (fns[i].evalfn) (ret, tmp, default_rmode);
	  return;
	}
    }

  if (ch == '(')
    {
      next;
      mpfr_expr(ret);
      if (ch == ')')
	next;
      else
	new_error ("mpfr_factor 3");
	//error ();
    }
  else
    mpfr_number(ret);
  if (ch == '!')
    {
      // unsigned long n;
      mpfr_t n;

      mpfr_init2(n, mpfr_get_prec(ret));

      mpfr_floor(tmp, ret);
      if (mpfr_cmp(tmp, ret) != 0)
	//error ();
	new_error ("mpfr_factor 4");
      mpfr_set_ui(ret, 1UL, default_rmode);
      for (mpfr_set(n, ret, default_rmode); mpfr_cmp_ui(n, 1) > 0; mpfr_sub_ui(n, n, 1UL, default_rmode))
	mpfr_mul(ret, ret, n, default_rmode);
      next;

      mpfr_clear(n);
    }

   mpfr_clear(tmp);
 // printf("mpfr_factor ended!\n"); fflush(stdout);
}

void mpfr_strtod (mpfr_t ret, char *head, char **next_head)
{
	char tmp_str[4096], *ch; // 
	long int i, j, exp_flag = 0;

 // printf("mpfr_strtod!\n"); fflush(stdout);

	ch = head;
	tmp_str[0] = '+';

	for(j = 0; j < 2048 - 1; j++)
	{
		if(ch[j] == '\0')
			break;

		if(ch[j] == '+' || ch[j] == '-')
		{
			tmp_str[0] = ch[j];
			break;
		}
		else if((ch[j] != ' ') && (ch[j] != '\t'))
			break;
	}

	for(i = 1; i < 4096 - 1; i++, j++)
	{
		if(ch[j] == '\0')
			break;

		if (((ch[j] >= '0') && (ch[j] <= '9')) || (ch[j] == '.'))
			tmp_str[i] = ch[j];
		else
		{
			if(exp_flag > 1)
				break;
			else if(exp_flag == 1)
			{
				if(ch[j] == '+' || ch[j] == '-')
				{
					tmp_str[i] = ch[j];
					exp_flag += 1;
				}
			}
			if((ch[j] == ' ') || (ch[j] == '\t'))
				continue;
			else if((ch[j] == 'E') || (ch[j] == 'e') || (ch[j] == 'Q') || (ch[j] == 'q') || (ch[j] == 'D') || (ch[j] == 'd'))
			{
				tmp_str[i] = ch[j];
				exp_flag += 1;
			}
			else
				break;
		}
	}
	tmp_str[i] = '\0';

//	printf("tmp_str: %s\n", tmp_str);
	mpfr_set_str(ret, tmp_str, 10, default_rmode);

	if(next_head != NULL)
		*(next_head) = &ch[j];
  //printf("mpfr_strtod ended!\n"); fflush(stdout);
}

void mpfr_number(mpfr_t ret)
{
  mpfr_t tmp;
  char *endp;

 // printf("mpfr_number!\n"); fflush(stdout);

  mpfr_init2(tmp, mpfr_get_prec(ret));

  /* Pi */
  if (strncmp ("pi", bp - 1, 2) == 0 && ! isalnum (bp[1]))
    {
      bp += 2 - 1;
      next;
      // return 3.1415926535897932384626433832795;
      mpfr_const_pi(ret, default_rmode);
      return;
    }
  /* Euler constant */
  if (strncmp ("eu", bp - 1, 2) == 0 && ! isalnum (bp[1]))
    {
      bp += 2 - 1;
      next;
      mpfr_const_euler(ret, default_rmode);
      return;
    }
  if (ch == '$')
    {
      if (! previous_result_valid_flag)
	//error ();
	new_error ("mpfr_number 1");
      next;
      // return previous_result;
      mpfr_set(ret, previous_result, default_rmode);
      return;
    }
//  printf("ch = %c\n", ch);
  if (ch != '.' && (ch < '0' || ch > '9'))
    // error ();
   new_error ("mpfr_number 2");
  mpfr_strtod (ret, bp - 1, &endp); //? mpfr_set_str
//  printf("%c, %c\n", *(bp - 1), *endp); //? mpfr_set_str
  if (endp == bp - 1)
    //error ();
    new_error ("mpfr_number 3");
  bp = endp;
  next;

  mpfr_clear(tmp);
 // printf("mpfr_number ended!\n"); fflush(stdout);
}
#endif

main (argc, argv)
     int argc;
     char **argv;
{
  int nl_flag = 1;
  int prec = 128;

#ifdef USE_GMP
  mpfr_t tmp;

  while (argc >= 2)
    {
      if (!strcmp (argv[1], "-n"))
	{
	  nl_flag = 0;
	  argv++;
	  argc--;
	}
      else if (!strcmp (argv[1], "-prec"))
	{
	  prec = atoi (argv[2]);
	  argv += 2;
	  argc -= 2;
	}
      else if (!strcmp (argv[1], "-help") || !strcmp (argv[1], "-h"))
	{
          printf ("usage: %s [options] expr [expr ... ]\n", argv[0]);
	  printf ("   options:    -n      -- suppress newline\n");
	  printf ("               -prec n -- print n digits\n");
	  printf ("               -help   -- make demons drop from your nose\n");
	  exit (0);
	}
      else
	break;
    }

  if (argc >= 2)
    {
      int i;

      mpfr_init2(tmp, prec);
      mpfr_init2(previous_result, prec);

      for (i = 1; i < argc; i++)
	{
	  expr_start_p = argv[i];
	  expr_end_p = expr_end_p + strlen (expr_start_p);
	  bp = expr_start_p;
	  next;
	  if (setjmp (top) == 0)
	    {
	      int h, m;
	      mpfr_expr(tmp);
	      mpfr_out_str(stdout, 10, 0, tmp, default_rmode);
	      if (nl_flag)
		puts ("");
	      mpfr_set(previous_result, tmp, default_rmode);
	      previous_result_valid_flag = 1;
	    }
	}
    }
  else
    {
#define BUFSIZE 1024
      char buf[BUFSIZE];

      mpfr_init2(tmp, prec);
      mpfr_init2(previous_result, prec);

      for (;;)
	{
	  fputs ("eval> ", stdout);
	  bp = fgets (buf, BUFSIZE, stdin);
	  //printf("%s\n", bp);
	  //if (bp == NULL)
	  if (strncmp(bp, "\n", 1) == 0)
	    break;
	  next;
	  if (setjmp (top) == 0)
	    {
	      int h, m;
	      mpfr_expr(tmp);
              mpfr_out_str(stdout, 10, 0, tmp, default_rmode);
	      if (nl_flag)
		puts ("");
	      mpfr_set(previous_result, tmp, default_rmode);
	      previous_result_valid_flag = 1;
	    }
	}
    }

  mpfr_clear(tmp);
  mpfr_clear(previous_result);
#endif

  exit (0);
}
