/***********************************************/
/* Test Program of Algebraic solvers           */
/***********************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

//#define USE_GMP
//#define USE_MPFR
#include "bnc.h"
#include <fenv.h> // C99

/* get rounding mode */
int get_rm(void)
{
	int rounding_mode;

	rounding_mode = fegetround();
	switch(rounding_mode)
	{
		case FE_DOWNWARD:
			printf("RM mode\n");
			break;
		case FE_UPWARD:
			printf("RP mode\n");
			break;
		case FE_TONEAREST:
			printf("RN mode\n");
			break;
		case FE_TOWARDZERO:
			printf("RZ mode\n");
			break;
		default:
			printf("Unknown!!\n");
			break;
	}

	return rounding_mode;
}

/* set rounding mode */
void set_rm(int rounding_mode)
{
	switch(rounding_mode)
	{
		case FE_DOWNWARD:
			if(fesetround(rounding_mode) == 0)
				printf("RM mode\n");
			else
				get_rm();
			break;
		case FE_UPWARD:
			if(fesetround(rounding_mode) == 0)
				printf("RP mode\n");
			else
				get_rm();
			break;
		case FE_TONEAREST:
			if(fesetround(rounding_mode) == 0)
				printf("RN mode\n");
			else
				get_rm();
			break;
		case FE_TOWARDZERO:
			if(fesetround(rounding_mode) == 0)
				printf("RZ mode\n");
			else
				get_rm();
			break;
		default:
			printf("Unknown!!\n");
			break;
	}
}

double dmax_array(double array[], long int num)
{
	double ret;
	long int i;

	ret = array[0];
	for(i = 1; i < num; i++)
	{
		if(ret < array[i])
			ret = array[i];
	}

	return ret;
}

int dcompare(const void *a, const void *b)
{
	double da, db;

	da = *(double *)a;
	db = *(double *)b;
	
	if(da > db)
		return -1;
	else if(da == db)
		return 0;
	else
		return +1;
}

#ifdef USE_GMP
int mpfcompare(const void *a, const void *b)
{
	int flag;
	mpf_t da, db;

//	da = (mpf_ptr)a;
//	db = (mpf_ptr)b;

	flag = mpf_cmp((mpf_ptr)a, (mpf_ptr)b);
	if(flag > 0)
		return -1;
	else if(flag == 0)
		return 0;
	else
		return +1;
}
#endif


// return max(|a|, |b|, |c|)
double dabsmax3(double a, double b, double c)
{
	double ret, tmp;

	ret = fabs(a);
	tmp = fabs(b);
	if(ret < tmp) ret = tmp;
	tmp = fabs(c);
	if(ret < tmp) ret = tmp;

	return ret;
}

// error
int dquadratic_eq_err(double ans_re[2], double ans_im[2], double ans_re_err[2], double ans_im_err[2], double coef[3])
{
	double tmp_ans_re[3][2], tmp_ans_im[3][2];

	set_rm(FE_DOWNWARD);
		dquadratic_eq(tmp_ans_re[0], tmp_ans_im[0], coef);

	set_rm(FE_UPWARD);
		dquadratic_eq(tmp_ans_re[1], tmp_ans_im[1], coef);

	set_rm(FE_TOWARDZERO);
		dquadratic_eq(tmp_ans_re[2], tmp_ans_im[2], coef);

	// Default Rounding mode
	set_rm(FE_TONEAREST);
		dquadratic_eq(ans_re, ans_im, coef);

	// get error of ans_re[0]
	ans_re_err[0] = dabsmax3(
		ans_re[0] - tmp_ans_re[0][0],
		ans_re[0] - tmp_ans_re[1][0],
		ans_re[0] - tmp_ans_re[2][0]);
	// get error of ans_re[1]
	ans_re_err[1] = dabsmax3(
		ans_re[1] - tmp_ans_re[0][1],
		ans_re[1] - tmp_ans_re[1][1],
		ans_re[1] - tmp_ans_re[2][1]);
	// get error of ans_im[0]
	ans_im_err[0] = dabsmax3(
		ans_im[0] - tmp_ans_im[0][0],
		ans_im[0] - tmp_ans_im[1][0],
		ans_im[0] - tmp_ans_im[2][0]);
	// get error of ans_im[1]
	ans_im_err[1] = dabsmax3(
		ans_im[1] - tmp_ans_im[0][1],
		ans_im[1] - tmp_ans_im[1][1],
		ans_im[1] - tmp_ans_im[2][1]);
}

#ifdef USE_GMP

// error
#define ADD_DPREC 10 // 10 decimal digits 
int mpfquadratic_eq_err_old(mpf_t ans_re[2], mpf_t ans_im[2], mpf_t ans_re_err[2], mpf_t ans_im_err[2], mpf_t coef[3])
{
	unsigned long prec, prec_s, prec_l, tmp_prec;
	mpf_t tmp_ans_re_l[2], tmp_ans_im_l[2];

	/* get default precision in bits */
	prec = mpf_get_prec(ans_re[0]);
	tmp_prec = mpf_get_prec(ans_re[1]);	if(tmp_prec > prec) prec = tmp_prec;
	tmp_prec = mpf_get_prec(ans_im[0]);	if(tmp_prec > prec) prec = tmp_prec;
	tmp_prec = mpf_get_prec(ans_im[1]);	if(tmp_prec > prec) prec = tmp_prec;

	/* set short prec and long prec */
	prec_s = prec;
	prec_l = prec + (unsigned long)ceil(ADD_DPREC/log10(2.0));

	/* initialize in long precision */
	mpf_init2(tmp_ans_re_l[0], prec_l);	mpf_init2(tmp_ans_re_l[1], prec_l);
	mpf_init2(tmp_ans_im_l[0], prec_l);	mpf_init2(tmp_ans_im_l[1], prec_l);

	/* get answers in short precision */
	mpfquadratic_eq(ans_re, ans_im, coef);

	/* get answers in long precision */
	mpfquadratic_eq(tmp_ans_re_l, tmp_ans_im_l, coef);

	// get error of ans_re[0, 1], ans_im[0, 1]
	mpf_sub(ans_re_err[0], ans_re[0], tmp_ans_re_l[0]);	mpf_abs(ans_re_err[0], ans_re_err[0]);
	mpf_sub(ans_re_err[1], ans_re[1], tmp_ans_re_l[1]);	mpf_abs(ans_re_err[1], ans_re_err[1]);
	mpf_sub(ans_im_err[0], ans_im[0], tmp_ans_im_l[0]);	mpf_abs(ans_im_err[0], ans_im_err[0]);
	mpf_sub(ans_im_err[1], ans_im[1], tmp_ans_im_l[1]);	mpf_abs(ans_im_err[1], ans_im_err[1]);

	/* free */
	mpf_clear(tmp_ans_re_l[0]);	mpf_clear(tmp_ans_re_l[1]);
	mpf_clear(tmp_ans_im_l[0]);	mpf_clear(tmp_ans_im_l[1]);
}

void mpf_get_coef(mpf_t coef[3], unsigned long coef_prec)
{
	mpf_t mpf_tmp;

	mpf_init2(mpf_tmp, coef_prec);
//	mpf_set_str(mpf_coef[2], "1", 10);
//	mpf_set_str(mpf_coef[1], "2", 10);
//	mpf_set_str(mpf_coef[0], "-3", 10);

	// dcoef[2] = exp(1.0);
	mpf_set_ui(mpf_tmp, 1UL);mpf_exp(coef[2], mpf_tmp);
	//dcoef[1] = -M_PI;
	mpf_pi(coef[1]); mpf_neg(coef[1], coef[1]);
	//dcoef[0] = 4.0 * log10(2.0);
	mpf_set_ui(mpf_tmp, 2UL);
	mpf_log10(coef[0], mpf_tmp); mpf_mul_ui(coef[0], coef[0], 4UL);

//	mpf_set_d(mpf_coef[2], dcoef[2]);
//	mpf_set_d(mpf_coef[1], dcoef[1]);
//	mpf_set_d(mpf_coef[0], dcoef[0]);

	mpf_clear(mpf_tmp);
}

// validated routine
#define MAX_PREC 16384
int mpfquadratic_eq_prec(mpf_t ans_re[2], mpf_t ans_im[2], mpf_t ans_re_err[2], mpf_t ans_im_err[2], void (*get_coef)(mpf_t coef[3], unsigned long coef_prec), unsigned long prec)
{
	unsigned long work_prec, prec_s, prec_l, prec_inc;
	long int flag;
	mpf_t tmp_ans_re_l[2], tmp_ans_im_l[2];
	mpf_t tmp_ans_re_s[2], tmp_ans_im_s[2];
	mpf_t coef[3]; // coef[2] * x^2 + coef[1] * x + coef[0] = 0.0;
	mpf_t rtol, atol, tmp[2];

	/* set short prec and long prec */
	prec_inc = lmax((unsigned long)(prec / 10), (unsigned long)ceil(ADD_DPREC/log10(2.0)));

	/* error */
//	mpf_init(ans_re_err[0]); mpf_init(ans_re_err[1]);
//	mpf_init(ans_im_err[0]); mpf_init(ans_im_err[1]);
	mpf_init(rtol); mpf_init(atol);
	mpf_init(tmp[0]); mpf_init(tmp[1]);

	/* rtol = 2^(-prec) */
	/* atol = 2^(-2*prec) */
	mpf_init_set_ui(rtol, 2UL); mpf_ui_div(rtol, 1UL, rtol); mpf_power(rtol, rtol, prec);
	mpf_init_set_ui(atol, 2UL); mpf_ui_div(atol, 1UL, atol); mpf_power(atol, atol, 2 * prec);

	/* get default precision in bits */
	for(work_prec = prec + prec_inc; work_prec <= MAX_PREC; work_prec += prec_inc)
	{
		prec_s = work_prec;
		prec_l = work_prec + prec_inc;
		printf("work_prec, prec_s, prec_l: %d, %d, %d\n", work_prec, prec_s, prec_l);

		/* initialize in long precision */
		mpf_init2(coef[0], prec_l); mpf_init2(coef[1], prec_l); mpf_init2(coef[2], prec_l);
		mpf_init2(tmp_ans_re_l[0], prec_l);	mpf_init2(tmp_ans_re_l[1], prec_l);
		mpf_init2(tmp_ans_im_l[0], prec_l);	mpf_init2(tmp_ans_im_l[1], prec_l);
		mpf_init2(tmp_ans_re_s[0], prec_s);	mpf_init2(tmp_ans_re_s[1], prec_s);
		mpf_init2(tmp_ans_im_s[0], prec_s);	mpf_init2(tmp_ans_im_s[1], prec_s);

		/* get coef[3] */
		get_coef(coef, prec_l);

		/* get answers in short and long precisions */
		mpfquadratic_eq(tmp_ans_re_s, tmp_ans_im_s, coef);
		mpfquadratic_eq(tmp_ans_re_l, tmp_ans_im_l, coef);

		/* get error of ans_re[0, 1], ans_im[0, 1] */
		mpf_sub(ans_re_err[0], tmp_ans_re_s[0], tmp_ans_re_l[0]); mpf_abs(ans_re_err[0], ans_re_err[0]);
		mpf_sub(ans_re_err[1], tmp_ans_re_s[1], tmp_ans_re_l[1]); mpf_abs(ans_re_err[1], ans_re_err[1]);
		mpf_sub(ans_im_err[0], tmp_ans_im_s[0], tmp_ans_im_l[0]); mpf_abs(ans_im_err[0], ans_im_err[0]);
		mpf_sub(ans_im_err[1], tmp_ans_im_s[1], tmp_ans_im_l[1]); mpf_abs(ans_im_err[1], ans_im_err[1]);

		/* check precision */
		flag = 0;
		mpf_mul(tmp[0], rtol, tmp_ans_re_l[0]); mpf_abs(tmp[0], tmp[0]);
		if((mpf_cmp(ans_re_err[0], tmp[0]) <= 0) || (mpf_cmp(ans_re_err[0], atol) <= 0))
			flag++;
		mpf_mul(tmp[0], rtol, tmp_ans_re_l[1]); mpf_abs(tmp[0], tmp[0]);
		if((mpf_cmp(ans_re_err[1], tmp[0]) <= 0) || (mpf_cmp(ans_re_err[1], atol) <= 0))
			flag++;
		mpf_mul(tmp[0], rtol, tmp_ans_im_l[0]); mpf_abs(tmp[0], tmp[0]);
		if((mpf_cmp(ans_im_err[0], tmp[0]) <= 0) || (mpf_cmp(ans_im_err[0], atol) <= 0))
			flag++;
		mpf_mul(tmp[0], rtol, tmp_ans_im_l[1]); mpf_abs(tmp[0], tmp[0]);
		if((mpf_cmp(ans_im_err[1], tmp[0]) <= 0) || (mpf_cmp(ans_im_err[1], atol) <= 0))
			flag++;

		/* Set */
		if(flag >= 4)
		{
			mpf_set(ans_re[0], tmp_ans_re_l[0]);
			mpf_set(ans_re[1], tmp_ans_re_l[1]);
			mpf_set(ans_im[0], tmp_ans_im_l[0]);
			mpf_set(ans_im[1], tmp_ans_im_l[1]);
		}

		/* free */
		mpf_clear(coef[0]); mpf_clear(coef[1]); mpf_clear(coef[2]);
		mpf_clear(tmp_ans_re_l[0]);	mpf_clear(tmp_ans_re_l[1]);
		mpf_clear(tmp_ans_im_l[0]);	mpf_clear(tmp_ans_im_l[1]);
		mpf_clear(tmp_ans_re_s[0]);	mpf_clear(tmp_ans_re_s[1]);
		mpf_clear(tmp_ans_im_s[0]);	mpf_clear(tmp_ans_im_s[1]);

		/* stop ! */
		if(flag >= 4)
			break;

	}

	/* free error */
//	mpf_clear(ans_re_err[0]); mpf_clear(ans_re_err[1]);
//	mpf_clear(ans_im_err[0]); mpf_clear(ans_im_err[1]);
	mpf_clear(rtol); mpf_clear(atol);
	mpf_clear(tmp[0]); mpf_clear(tmp[1]);
}
#endif

main()
{
	double dcoef[3]; // dcoef[2] * x^2 + dcoef[1] * x + dcoef[0] = 0.0;
	double dans_re[2], dans_im[2]; // ans = dans_re[0 or 1] + dans_im[0 or 1] * sqrt(-1)
	double dans_re_err[2], dans_im_err[2];
#ifdef USE_GMP
	mpf_t mpf_coef[3]; // mpf_coef[2] * x^2 + mpf_coef[1] * x + mpf_coef[0] = 0.0;
	mpf_t mpf_ans_re[2], mpf_ans_im[2]; // ans = mpf_ans_re[0 or 1] + mpf_ans_im[0 or 1] * sqrt(-1)
	mpf_t mpf_ans_re_err[2], mpf_ans_im_err[2], mpf_tmp;
#endif

	dcoef[2] = exp(1.0);
	dcoef[1] = -M_PI;
	dcoef[0] = 4.0 * log10(2.0);

	// solve quadratic equation
	dquadratic_eq_err(dans_re, dans_im, dans_re_err, dans_im_err, dcoef);

	printf("ans[0] = %25.17e, %25.17e\n", dans_re[0], dans_im[0]);
	printf("ans[1] = %25.17e, %25.17e\n", dans_re[1], dans_im[1]);
	printf("err[0] = %25.17e, %25.17e\n", dans_re_err[0], dans_im_err[0]);
	printf("err[1] = %25.17e, %25.17e\n", dans_re_err[1], dans_im_err[1]);

#ifdef USE_GMP
#define DPREC 50

	set_bnc_default_prec_decimal(DPREC);

	mpf_init(mpf_tmp);
	mpf_init(mpf_coef[2]);
	mpf_init(mpf_coef[1]);
	mpf_init(mpf_coef[0]);
	mpf_init(mpf_ans_re[0]); mpf_init(mpf_ans_im[0]);
	mpf_init(mpf_ans_re[1]); mpf_init(mpf_ans_im[1]);
	mpf_init(mpf_ans_re_err[0]); mpf_init(mpf_ans_im_err[0]);
	mpf_init(mpf_ans_re_err[1]); mpf_init(mpf_ans_im_err[1]);

//	mpf_set_str(mpf_coef[2], "1", 10);
//	mpf_set_str(mpf_coef[1], "2", 10);
//	mpf_set_str(mpf_coef[0], "-3", 10);

	// dcoef[2] = exp(1.0);
	mpf_set_ui(mpf_tmp, 1UL);mpf_exp(mpf_coef[2], mpf_tmp);
	//dcoef[1] = -M_PI;
	mpf_pi(mpf_coef[1]); mpf_neg(mpf_coef[1], mpf_coef[1]);
	//dcoef[0] = 4.0 * log10(2.0);
	mpf_set_ui(mpf_tmp, 2UL);
	mpf_log10(mpf_coef[0], mpf_tmp); mpf_mul_ui(mpf_coef[0], mpf_coef[0], 4UL);

//	mpf_set_d(mpf_coef[2], dcoef[2]);
//	mpf_set_d(mpf_coef[1], dcoef[1]);
//	mpf_set_d(mpf_coef[0], dcoef[0]);

	// solve quadratic equation
	//mpfquadratic_eq_err(mpf_ans_re, mpf_ans_im, mpf_ans_re_err, mpf_ans_im_err, mpf_coef);
	mpfquadratic_eq_prec(mpf_ans_re, mpf_ans_im, mpf_ans_re_err, mpf_ans_im_err, mpf_get_coef, get_bnc_default_prec());

	printf("ans[0] ="); mpf_out_str(stdout, 10, 0, mpf_ans_re[0]);
	printf(", "); mpf_out_str(stdout, 10, 0, mpf_ans_im[0]); printf("\n");
	printf("ans[1] ="); mpf_out_str(stdout, 10, 0, mpf_ans_re[1]);
	printf(", "); mpf_out_str(stdout, 10, 0, mpf_ans_im[1]); printf("\n");
	printf("ans[0]_err ="); mpf_out_str(stdout, 10, 3, mpf_ans_re_err[0]);
	printf(", "); mpf_out_str(stdout, 10, 3, mpf_ans_im_err[0]); printf("\n");
	printf("ans[1]_err ="); mpf_out_str(stdout, 10, 3, mpf_ans_re_err[1]);
	printf(", "); mpf_out_str(stdout, 10, 3, mpf_ans_im_err[1]); printf("\n");

	mpf_clear(mpf_tmp);
	mpf_clear(mpf_coef[2]);
	mpf_clear(mpf_coef[1]);
	mpf_clear(mpf_coef[0]);
	mpf_clear(mpf_ans_re[0]); mpf_clear(mpf_ans_im[0]);
	mpf_clear(mpf_ans_re[1]); mpf_clear(mpf_ans_im[1]);
	mpf_clear(mpf_ans_re_err[0]); mpf_clear(mpf_ans_im_err[0]);
	mpf_clear(mpf_ans_re_err[1]); mpf_clear(mpf_ans_im_err[1]);
#endif
}
