/**********************************************/
/* test_krylov.c:                             */
/* Copyright (C) 2004- Tomonori Kouya         */
/*                                            */
/* This library is free software; you can re- */
/* distribute it and/or modify it under the   */
/* terms of the GNU Lesser General Public     */
/* License as published by the Free Software  */
/* Foundation; either version 2.1 of the      */
/* License, or (at your option) any later     */
/* version.                                   */
/*                                            */
/* This library is distributed in the hope    */
/* that it will be useful, but WITHOUT ANY    */
/* WARRANTY; without even the implied         */
/* warranty of MERCHANTABILITY or FITNESS FOR */
/* A PARTICULAR PURPOSE.  See the GNU Lesser  */
/* General Public License for more details.   */
/**********************************************/
/****************************************************/
/* Test Program for *BiCG, *CGS, *BiCGSTAB, *GPBiCG */
/****************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bnc.h"

void get_dproblem(DMatrix a, DVector b, DVector ans)
{
	long int i, j, k;
	double tmp;

	/* Lotkin Matrix */
	lotkin_dmatrix(a, a->row_dim);

	/* Answer */
	for(i = 0; i < ans->dim; i++)
		set_dvector_i(ans, i, (double)i);

	/* Make constant vector */
	mul_dmatrix_dvec(b, a, ans);
}

#ifdef USE_GMP
void get_mpfproblem(MPFMatrix a, MPFVector b, MPFVector ans)
{
	long int i, j, k;
	mpf_t tmp;

	mpf_init2(tmp, prec_mpfvector(ans));

	/* Lotkin Matrix */
	lotkin_mpfmatrix(a, a->row_dim);

	/* Answer */
	for(i = 0; i < ans->dim; i++)
	{
		mpf_set_si(tmp, i);
		set_mpfvector_i(ans, i, tmp);
	}

	/* Make constant vector */
	mul_mpfmatrix_mpfvec(b, a, ans);
}
#endif

#define DIM 100

main()
{
	DMatrix da;
	DVector db, dx, dans;
	double start, dtime[4];
#ifdef USE_GMP
	MPFMatrix mpfa;
	MPFVector mpfb, mpfx, mpfans;
	mpf_t reps, aeps;
	MPFMatrix mpfa2;
	MPFVector mpfb2, mpfx2, mpfans2;
	mpf_t reps2, aeps2;
	MPFMatrix mpfa3;
	MPFVector mpfb3, mpfx3, mpfans3;
	mpf_t reps3, aeps3;
	long int itimes_mpf[3][4];
	double mpftime[3][4];
#endif
	long int itimes_d[4];
	long int i, j;

/* Double */
	/* initialize */
	da = init_dmatrix(DIM, DIM);
	db = init_dvector(DIM);
	dx = init_dvector(DIM);
	dans = init_dvector(DIM);

	/* get problem */
	get_dproblem(da, db, dans);

//	print_dmatrix(da);

	/* run DBiCG */
	start = get_secv();
	itimes_d[0] = DBiCG(dx, da, db, 1.0e-13, 1.0e-99, DIM * 5);
	dtime[0] = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
		printf("%5ld %25.17e %25.17e\n", i, get_dvector_i(dx, i), get_dvector_i(dans, i));

	/* get problem */
	get_dproblem(da, db, dans);

//	print_dmatrix(da);

	/* run DCGS */
	start = get_secv();
	itimes_d[1] = DCGS(dx, da, db, 1.0e-13, 1.0e-99, DIM * 5);
	dtime[1] = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
		printf("%5ld %25.17e %25.17e\n", i, get_dvector_i(dx, i), get_dvector_i(dans, i));

	/* get problem */
	get_dproblem(da, db, dans);

//	print_dmatrix(da);

	/* run DBiCGSTAB */
	start = get_secv();
	itimes_d[2] = DBiCGSTAB(dx, da, db, 1.0e-13, 1.0e-99, DIM * 5);
	dtime[2] = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
		printf("%5ld %25.17e %25.17e\n", i, get_dvector_i(dx, i), get_dvector_i(dans, i));

	/* get problem */
	get_dproblem(da, db, dans);

//	print_dmatrix(da);

	/* run DGPBiCG */
	start = get_secv();
	itimes_d[3] = DGPBiCG(dx, da, db, 1.0e-13, 1.0e-99, DIM * 5);
	dtime[3] = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
		printf("%5ld %25.17e %25.17e\n", i, get_dvector_i(dx, i), get_dvector_i(dans, i));
		
	/* end */
	free_dmatrix(da);
	free_dvector(db);
	free_dvector(dx);
	free_dvector(dans);

#ifdef USE_GMP
/* MPF */
	set_bnc_default_prec(128);
	/* initialize */
	mpf_init(reps);
	mpf_init2(reps2, 256);
	mpf_init2(reps3, 512);
	mpf_init(aeps);
	mpf_init2(aeps2, 256);
	mpf_init2(aeps3, 512);
	mpfa  = init_mpfmatrix(DIM, DIM);
	mpfa2 = init2_mpfmatrix(DIM, DIM, 256);
	mpfa3 = init2_mpfmatrix(DIM, DIM, 512);
	mpfb  = init_mpfvector(DIM);
	mpfb2 = init2_mpfvector(DIM, 256);
	mpfb3 = init2_mpfvector(DIM, 512);
	mpfx  = init_mpfvector(DIM);
	mpfx2 = init2_mpfvector(DIM, 256);
	mpfx3 = init2_mpfvector(DIM, 512);
	mpfans  = init_mpfvector(DIM);
	mpfans2 = init2_mpfvector(DIM, 256);
	mpfans3 = init2_mpfvector(DIM, 512);

//	print_mpfmatrix(mpfa);
//	print_mpfmatrix(mpfa2);
//	print_mpfmatrix(mpfa3);

	mpf_set_d(reps, 1.0e-20);
	mpf_set_d(reps2, 1.0e-20);
	mpf_set_d(reps3, 1.0e-20);
	mpf_set_d(aeps, 1.0e-50);
	mpf_set_d(aeps2, 1.0e-50);
	mpf_set_d(aeps3, 1.0e-50);

//	goto MPFGPBiCG;
MPFBiCG:

	/* run MPFBiCG */
	get_mpfproblem(mpfa, mpfb, mpfans);
	start = get_secv();
	itimes_mpf[0][0]  = MPFBiCG(mpfx, mpfa, mpfb, reps, aeps, DIM * 5);
	mpftime[0][0] = get_secv() - start;

	get_mpfproblem(mpfa2, mpfb2, mpfans2);
	start = get_secv();
	itimes_mpf[1][0] = MPFBiCG(mpfx2, mpfa2, mpfb2, reps2, aeps2, DIM * 5);
	mpftime[1][0] = get_secv() - start;

	get_mpfproblem(mpfa3, mpfb3, mpfans3);
	start = get_secv();
	itimes_mpf[2][0] = MPFBiCG(mpfx3, mpfa3, mpfb3, reps3, aeps3, DIM * 5);
	mpftime[2][0] = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
	{
		printf("%5ld ", i);
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx2, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx3, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfans, i));
		printf("\n");
	}

	/* run MPFCGS */
	get_mpfproblem(mpfa, mpfb, mpfans);
	start = get_secv();
	itimes_mpf[0][1]  = MPFCGS(mpfx, mpfa, mpfb, reps, aeps, DIM * 5);
	mpftime[0][1] = get_secv() - start;

	get_mpfproblem(mpfa2, mpfb2, mpfans2);
	start = get_secv();
	itimes_mpf[1][1] = MPFCGS(mpfx2, mpfa2, mpfb2, reps2, aeps2, DIM * 5);
	mpftime[1][1] = get_secv() - start;

	get_mpfproblem(mpfa3, mpfb3, mpfans3);
	start = get_secv();
	itimes_mpf[2][1] = MPFCGS(mpfx3, mpfa3, mpfb3, reps3, aeps3, DIM * 5);
	mpftime[2][1] = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
	{
		printf("%5ld ", i);
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx2, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx3, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfans, i));
		printf("\n");
	}

MPFBiCGSTAB:

	/* run MPFBiCGSTAB */
	get_mpfproblem(mpfa, mpfb, mpfans);
	start = get_secv();
	itimes_mpf[0][2]  = MPFBiCGSTAB(mpfx, mpfa, mpfb, reps, aeps, DIM * 5);
	mpftime[0][2] = get_secv() - start;

	get_mpfproblem(mpfa2, mpfb2, mpfans2);
	start = get_secv();
	itimes_mpf[1][2] = MPFBiCGSTAB(mpfx2, mpfa2, mpfb2, reps2, aeps2, DIM * 5);
	mpftime[1][2] = get_secv() - start;

	get_mpfproblem(mpfa3, mpfb3, mpfans3);
	start = get_secv();
	itimes_mpf[2][2] = MPFBiCGSTAB(mpfx3, mpfa3, mpfb3, reps3, aeps3, DIM * 5);
	mpftime[2][2] = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
	{
		printf("%5ld ", i);
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx2, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx3, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfans, i));
		printf("\n");
	}

MPFGPBiCG:

	/* run MPFGPBiCG */
	get_mpfproblem(mpfa, mpfb, mpfans);
	start = get_secv();
	itimes_mpf[0][3]  = MPFGPBiCG(mpfx, mpfa, mpfb, reps, aeps, DIM * 5);
	mpftime[0][3] = get_secv() - start;

	get_mpfproblem(mpfa2, mpfb2, mpfans2);
	start = get_secv();
	itimes_mpf[1][3] = MPFGPBiCG(mpfx2, mpfa2, mpfb2, reps2, aeps2, DIM * 5);
	mpftime[1][3] = get_secv() - start;

	get_mpfproblem(mpfa3, mpfb3, mpfans3);
	start = get_secv();
	itimes_mpf[2][3] = MPFGPBiCG(mpfx3, mpfa3, mpfb3, reps3, aeps3, DIM * 5);
	mpftime[2][3] = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
	{
		printf("%5ld ", i);
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx2, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx3, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfans, i));
		printf("\n");
	}

	/* end */
	mpf_clear(reps); mpf_clear(aeps);
	mpf_clear(reps2); mpf_clear(aeps2);
	mpf_clear(reps3); mpf_clear(aeps3);
	free_mpfmatrix(mpfa);
	free_mpfmatrix(mpfa2);
	free_mpfmatrix(mpfa3);
	free_mpfvector(mpfb);
	free_mpfvector(mpfb2);
	free_mpfvector(mpfb3);
	free_mpfvector(mpfx);
	free_mpfvector(mpfx2);
	free_mpfvector(mpfx3);
	free_mpfvector(mpfans);
	free_mpfvector(mpfans2);
	free_mpfvector(mpfans3);
#endif

	/* print itimes */
	printf("Iterative Times\n");
	printf("double(BiCG)    : %ld(%f)\n", itimes_d[0], dtime[0]);
	printf("double(CGS)     : %ld(%f)\n", itimes_d[1], dtime[1]);
	printf("double(BiCGSTAB): %ld(%f)\n", itimes_d[2], dtime[2]);
	printf("double(GPBiCG)  : %ld(%f)\n", itimes_d[3], dtime[3]);
#ifdef USE_GMP
#ifdef USE_MPFR
	printf("mpfr_t(128,BiCG)    : %ld(%f)\n", itimes_mpf[0][0], mpftime[0][0]);
	printf("mpfr_t(128,CGS)     : %ld(%f)\n", itimes_mpf[0][1], mpftime[0][1]);
	printf("mpfr_t(128,BiCGSTAB): %ld(%f)\n", itimes_mpf[0][2], mpftime[0][2]);
	printf("mpfr_t(128,GPBiCG)  : %ld(%f)\n", itimes_mpf[0][3], mpftime[0][3]);
	printf("mpfr_t(256,BiCG)    : %ld(%f)\n", itimes_mpf[1][0], mpftime[1][0]);
	printf("mpfr_t(256,CGS)     : %ld(%f)\n", itimes_mpf[1][1], mpftime[1][1]);
	printf("mpfr_t(256,BiCGSTAB): %ld(%f)\n", itimes_mpf[1][2], mpftime[1][2]);
	printf("mpfr_t(256,GPBiCG)  : %ld(%f)\n", itimes_mpf[1][3], mpftime[1][3]);
	printf("mpfr_t(512,BiCG)    : %ld(%f)\n", itimes_mpf[2][0], mpftime[2][0]);
	printf("mpfr_t(512,CGS)     : %ld(%f)\n", itimes_mpf[2][1], mpftime[2][1]);
	printf("mpfr_t(512,BiCGSTAB): %ld(%f)\n", itimes_mpf[2][2], mpftime[2][2]);
	printf("mpfr_t(512,GPBiCG)  : %ld(%f)\n", itimes_mpf[2][3], mpftime[2][3]);
#else
	printf("mpf_t(128): %ld(%f)\n", itimes_mpf, mpftime[0]);
	printf("mpf_t(256): %ld(%f)\n", itimes_mpf2, mpftime[1]);
	printf("mpf_t(512): %ld(%f)\n", itimes_mpf3, mpftime[2]);
#endif
#endif
}

