/**********************************************/
/* test_lu.c:                                 */
/* Copyright (C) 2003 Tomonori Kouya          */
/*                                            */
/* This library is free software; you can re- */
/* distribute it and/or modify it under the   */
/* terms of the GNU Lesser General Public     */
/* License as published by the Free Software  */
/* Foundation; either version 2.1 of the      */
/* License, or (at your option) any later     */
/* version.                                   */
/*                                            */
/* This library is distributed in the hope    */
/* that it will be useful, but WITHOUT ANY    */
/* WARRANTY; without even the implied         */
/* warranty of MERCHANTABILITY or FITNESS FOR */
/* A PARTICULAR PURPOSE.  See the GNU Lesser  */
/* General Public License for more details.   */
/**********************************************/
/*************************************************/
/* Test Program for                              */
/* FLUdecomp/SolveFLS, FLUdecompP/SolveFLUP, FLUdecompC/SolveFLSC */
/* DLUdecomp/SolveDLS, DLUdecompP/SolveDLUP, DLUdecompC/SolveDLSC */
/* MPFLUdecomp/SolveMPFLS, MPFLUdecompP/SolveMPFLUP, MPFLUdecompC/SolveMPFLSC*/
/*************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bnc.h"

void get_fproblem(FMatrix a, FVector b, FVector ans)
{
	long int i, j, k;
	float tmp;

	/* Lotkin Matrix */
	for(i = 0; i < a->col_dim; i++)
		set_fmatrix_ij(a, 0, i, 1.0);
	for(i = 1; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
			set_fmatrix_ij(a, i, j, 1.0 / (i + j + 1));
	}

	/* Answer */
	for(i = 0; i < ans->dim; i++)
		set_fvector_i(ans, i, (float)i);

	/* Make constant vector */
	mul_fmatrix_fvec(b, a, ans);
}

void get_dproblem(DMatrix a, DVector b, DVector ans)
{
	long int i, j, k;
	double tmp;

	/* Lotkin Matrix */
	for(i = 0; i < a->col_dim; i++)
		set_dmatrix_ij(a, 0, i, 1.0);
	for(i = 1; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
			set_dmatrix_ij(a, i, j, 1.0 / (i + j + 1));
	}

	/* Answer */
	for(i = 0; i < ans->dim; i++)
		set_dvector_i(ans, i, (double)i);

	/* Make constant vector */
	mul_dmatrix_dvec(b, a, ans);
}

#ifdef USE_GMP
void get_mpfproblem(MPFMatrix a, MPFVector b, MPFVector ans)
{
	long int i, j, k;
	mpf_t tmp;

	mpf_init(tmp);

	/* Lotkin Matrix */
	for(i = 0; i < a->col_dim; i++)
		set_mpfmatrix_ij_d(a, 0, i, 1.0);
	for(i = 1; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
		{
			mpf_set_ui(tmp, 1UL);
			mpf_div_ui(tmp, tmp, (unsigned long)(i + j + 1));
			set_mpfmatrix_ij(a, i, j, tmp);
		}
	}


	/* Answer */
	for(i = 0; i < ans->dim; i++)
	{
		mpf_set_si(tmp, i);
		set_mpfvector_i(ans, i, tmp);
	}

	/* Make constant vector */
	mul_mpfmatrix_mpfvec(b, a, ans);
}
#endif

#define DIM 5

main()
{
	FMatrix fa;
	FVector fb, fx, fans;
	DMatrix da;
	DVector db, dx, dans;
#ifdef USE_GMP
	MPFMatrix mpfa;
	MPFVector mpfb, mpfx, mpfans;
	mpf_t reps, aeps;
#endif
	long int ret_f, ret_d, ret_mpf;
	long int row_ch[DIM], col_ch[DIM];
	long int i, j;

/* Float */
fstart:
	/* initialize */
	fa = init_fmatrix(DIM, DIM);
	fb = init_fvector(DIM);
	fx = init_fvector(DIM);
	fans = init_fvector(DIM);

	/* get problem */
	get_fproblem(fa, fb, fans);

//	print_fmatrix(fa);

	/* run FLUdecomp & SolveFLS */
	// ret_f = FLUdecomp(fa);
	// ret_f = FLUdecompP(fa, row_ch);
	 ret_f = FLUdecompC(fa, row_ch, col_ch);
	// ret_f = SolveFLS(fx, fa, fb);
	// ret_f = SolveFLSP(fx, fa, fb, row_ch);
	 ret_f = SolveFLSC(fx, fa, fb, row_ch, col_ch);

	/* print */
	for(i = 0; i < DIM; i++)
		printf("%5ld %15.7e %15.7e\n", i, get_fvector_i(fx, i), get_fvector_i(fans, i));
		
	/* end */
	free_fmatrix(fa);
	free_fvector(fb);
	free_fvector(fx);
	free_fvector(fans);
//goto end;

/* Double */
dstart:
	/* initialize */
	da = init_dmatrix(DIM, DIM);
	db = init_dvector(DIM);
	dx = init_dvector(DIM);
	dans = init_dvector(DIM);

	/* get problem */
	get_dproblem(da, db, dans);

//	print_dmatrix(da);

	/* run DLUdecomp & SolveDLS */
	// ret_d = DLUdecomp(da);
	// ret_d = DLUdecompP(da, row_ch);
	 ret_d = DLUdecompC(da, row_ch, col_ch);
	// ret_d = SolveDLS(dx, da, db);
	// ret_d = SolveDLSP(dx, da, db, row_ch);
	 ret_d = SolveDLSC(dx, da, db, row_ch, col_ch);

	/* print */
	printf("  i    row_ch[i]    col_ch[i]\n");
	for(i = 0; i < DIM; i++)
		printf("%5ld %25.17e %25.17e\n", i, get_dvector_i(dx, i), get_dvector_i(dans, i));
		
	/* end */
	free_dmatrix(da);
	free_dvector(db);
	free_dvector(dx);
	free_dvector(dans);

//	goto end;

#ifdef USE_GMP
/* MPF */
mpfstart:
	set_bnc_default_prec(256);
	/* initialize */
	mpf_init(reps); mpf_init(aeps);
	mpfa = init_mpfmatrix(DIM, DIM);
//	mpfa = init2_mpfmatrix(DIM, DIM, 256);
	mpfb = init_mpfvector(DIM);
//	mpfb = init2_mpfvector(DIM, 256);
	mpfx = init_mpfvector(DIM);
//	mpfx = init2_mpfvector(DIM, 256);
	mpfans = init_mpfvector(DIM);

	/* get problem */
	get_mpfproblem(mpfa, mpfb, mpfans);

//	print_mpfmatrix(mpfa);

	/* run MPFLUdecomp & SolveMPFLS */
	// ret_mpf = MPFLUdecomp(mpfa);
	// ret_mpf = MPFLUdecompP(mpfa, row_ch);
	 ret_mpf = MPFLUdecompC(mpfa, row_ch, col_ch);
	// ret_mpf = SolveMPFLS(mpfx, mpfa, mpfb);
	// ret_mpf = SolveMPFLSP(mpfx, mpfa, mpfb, row_ch);
	 ret_mpf = SolveMPFLSC(mpfx, mpfa, mpfb, row_ch, col_ch);

	/* print */
	for(i = 0; i < DIM; i++)
	{
		printf("%5ld ", i);
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfans, i));
		printf("\n");
	}

	/* end */
	mpf_clear(reps); mpf_clear(aeps);
	free_mpfmatrix(mpfa);
	free_mpfvector(mpfb);
	free_mpfvector(mpfx);
	free_mpfvector(mpfans);
#endif

	/* print itimes */
end:
	return 0;

}

