/**********************************************/
/* test_lu.c:                                 */
/* Copyright (C) 2003 Tomonori Kouya          */
/*                                            */
/* This library is free software; you can re- */
/* distribute it and/or modify it under the   */
/* terms of the GNU Lesser General Public     */
/* License as published by the Free Software  */
/* Foundation; either version 2.1 of the      */
/* License, or (at your option) any later     */
/* version.                                   */
/*                                            */
/* This library is distributed in the hope    */
/* that it will be useful, but WITHOUT ANY    */
/* WARRANTY; without even the implied         */
/* warranty of MERCHANTABILITY or FITNESS FOR */
/* A PARTICULAR PURPOSE.  See the GNU Lesser  */
/* General Public License for more details.   */
/**********************************************/
/*************************************************/
/* Test Program for                              */
/* FLUdecomp/SolveFLS, FLUdecompP/SolveFLUP, FLUdecompC/SolveFLSC */
/* DLUdecomp/SolveDLS, DLUdecompP/SolveDLUP, DLUdecompC/SolveDLSC */
/* MPFLUdecomp/SolveMPFLS, MPFLUdecompP/SolveMPFLUP, MPFLUdecompC/SolveMPFLSC*/
/*************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bnc.h"

void get_dproblem(DMatrix a, DVector b, DVector ans)
{
	long int i, j, k;
	double tmp;

	/* Lotkin Matrix */
	for(i = 0; i < a->col_dim; i++)
	{
		set_dmatrix_ij(a, 0, i, 1.0);
	}
	for(i = 1; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
			set_dmatrix_ij(a, i, j, 1.0 / (i + j + 1));
	}

	/* Answer */
	for(i = 0; i < ans->dim; i++)
		set_dvector_i(ans, i, (double)i);

	/* Make constant vector */
	mul_dmatrix_dvec(b, a, ans);

}

#ifdef USE_GMP
void get_mpfproblem(MPFMatrix a, MPFVector b, MPFVector ans)
{
	long int i, j, k;
	mpf_t tmp;

	mpf_init(tmp);

	/* Lotkin Matrix */
	for(i = 0; i < a->col_dim; i++)
		set_mpfmatrix_ij_d(a, 0, i, 1.0);
	for(i = 1; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
		{
			mpf_set_ui(tmp, 1UL);
			mpf_div_ui(tmp, tmp, (unsigned long)(i + j + 1));
			set_mpfmatrix_ij(a, i, j, tmp);
		}
	}


	/* Answer */
	for(i = 0; i < ans->dim; i++)
	{
		mpf_set_si(tmp, i);
		set_mpfvector_i(ans, i, tmp);
	}

	/* Make constant vector */
	mul_mpfmatrix_mpfvec(b, a, ans);
}

void subst_mpfvector_dvec(MPFVector ret, DVector src)
{
	long int i;

	for(i = 0; i < ret->dim; i++)
		set_mpfvector_i_d(ret, i, get_dvector_i(src, i));
}

void subst_dvector_mpfvec(DVector ret, MPFVector src)
{
	long int i;

	for(i = 0; i < ret->dim; i++)
		set_dvector_i(ret, i, mpf_get_d(get_mpfvector_i(src, i)));
}

void subst_mpfmatrix_dmat(MPFMatrix ret, DMatrix src)
{
	long int i, j;

	for(i = 0; i < ret->row_dim; i++)
	{
		for(j = 0; j < ret->col_dim; j++)
			set_mpfmatrix_ij_d(ret, i, j, get_dmatrix_ij(src, i, j));
	}
}


#endif

#define DIM 10

main()
{
	DMatrix da;
	DVector db, dx, dans;
#ifdef USE_GMP
	MPFMatrix mpfa;
	MPFVector mpfb, mpfx, mpfans;
	mpf_t reps, aeps;
#endif
	long int ret_f, ret_d, ret_mpf;
	long int row_ch[DIM], col_ch[DIM];
	long int i, j;

/* Double */
dstart:
	/* initialize */
	da = init_dmatrix(DIM, DIM);
	db = init_dvector(DIM);
	dx = init_dvector(DIM);
	dans = init_dvector(DIM);

#ifdef USE_GMP
	set_bnc_default_prec(256);
	/* initialize */
	mpf_init(reps); mpf_init(aeps);
	mpfa = init_mpfmatrix(DIM, DIM);
//	mpfa = init2_mpfmatrix(DIM, DIM, 256);
	mpfb = init_mpfvector(DIM);
//	mpfb = init2_mpfvector(DIM, 256);
	mpfx = init_mpfvector(DIM);
//	mpfx = init2_mpfvector(DIM, 256);
	mpfans = init_mpfvector(DIM);

#endif
	/* get problem */
	get_dproblem(da, db, dans);

	print_dmatrix(da);
#ifdef USE_GMP
	subst_mpfmatrix_dmat(mpfa, da);
	subst_mpfvector_dvec(mpfb, db);
#endif

	/* run DLUdecomp & SolveDLS */
	// ret_d = DLUdecomp(da);
	// ret_d = DLUdecompP(da, row_ch);
	 ret_d = DLUdecompC(da, row_ch, col_ch);
	// ret_d = SolveDLS(dx, da, db);
	// ret_d = SolveDLSP(dx, da, db, row_ch);
	 ret_d = SolveDLSC(dx, da, db, row_ch, col_ch);
#ifdef USE_GMP
	subst_mpfvector_dvec(mpfans, dx);
	mul_mpfmatrix_mpfvec(mpfx, mpfa, mpfans); 
	sub_mpfvector(mpfans, mpfb, mpfx);// b - Ax
	ret_mpf = MPFLUdecompC(mpfa, row_ch, col_ch);
	ret_d = SolveMPFLSC(mpfx, mpfa, mpfb, row_ch, col_ch);
//	subst_dvector_mpfvec(db, mpfans);
//	ret_d = SolveDLSC(dans, da, db, row_ch, col_ch);
	subst_dvector_mpfvec(dans, mpfx);
	add_dvector(db, dans, dx);
#endif
	/* print */
	printf("  i    row_ch[i]    col_ch[i]\n");
	for(i = 0; i < DIM; i++)
		printf("%5ld %25.17e %25.17e %25.17e\n", i, get_dvector_i(dx, i), get_dvector_i(dans, i), get_dvector_i(db, i));
		
	/* end */
	free_dmatrix(da);
	free_dvector(db);
	free_dvector(dx);
	free_dvector(dans);

#ifdef USE_GMP
/* MPF */
mpfstart:

	/* get problem */
	get_mpfproblem(mpfa, mpfb, mpfans);

//	print_mpfmatrix(mpfa);

	/* run MPFLUdecomp & SolveMPFLS */
	 ret_mpf = MPFLUdecomp(mpfa);
	// ret_mpf = MPFLUdecompP(mpfa, row_ch);
	// ret_mpf = MPFLUdecompC(mpfa, row_ch, col_ch);
	 ret_mpf = SolveMPFLS(mpfx, mpfa, mpfb);
	// ret_mpf = SolveMPFLSP(mpfx, mpfa, mpfb, row_ch);
	// ret_mpf = SolveMPFLSC(mpfx, mpfa, mpfb, row_ch, col_ch);

	/* print */
	for(i = 0; i < DIM; i++)
	{
		printf("%5ld ", i);
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfans, i));
		printf("\n");
	}

	/* end */
	mpf_clear(reps); mpf_clear(aeps);
	free_mpfmatrix(mpfa);
	free_mpfvector(mpfb);
	free_mpfvector(mpfx);
	free_mpfvector(mpfans);
#endif

	/* print itimes */
end:

}

